#!/usr/bin/env python

import os
import tblock.rules
import tblock.exceptions
import tblock.config
import unittest
from nose.tools import raises

tblock.config.Path.HOSTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts')
tblock.config.Path.DATABASE = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'storage.sqlite')
tblock.config.Path.DB_LOCK = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', '.db-lock')
tblock.config.Path.TMP_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root')
tblock.config.Path.HOSTS_VERIFICATION = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'verification')
tblock.config.setup_database(tblock.config.Path.DATABASE)


class TestUserRules(unittest.TestCase):

    def test_user_rule_allow(self):
        self.assertEquals(
            tblock.rules.add_rules([['www.example.com', 'allow'], ['www.example.org', 'allow']], force=True),
            None
        )

    def test_user_rule_block(self):
        self.assertEquals(
            tblock.rules.add_rules([['www.example.com', 'block'], ['www.example.org', 'block']], force=True),
            None
        )

    def test_user_rule_redirect(self):
        self.assertEquals(
            tblock.rules.add_rules([
                ['www.example.com', 'redirect', '0.0.0.0'], ['www.example.org', 'redirect', '127.0.0.1']
            ], force=True),
            None
        )

    @raises(tblock.exceptions.InvalidBindAddress)
    def test_user_rule_redirect_fail(self):
        self.assertEquals(
            tblock.rules.add_rules([
                ['www.example.com', 'redirect'], ['www.example.org', 'redirect', '127.0.0.1']
            ], force=True),
            None
        )

    def test_user_rule_delete(self):
        try:
            os.remove(tblock.config.Path.DB_LOCK)
        except FileNotFoundError:
            pass
        self.assertEquals(
            tblock.rules.delete_rules(['www.example.com', 'www.example.org'], force=True),
            None
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import os
import sqlite3
import random

# External libraries
from colorama import init

# Compatibility with Windows
# See: https://codeberg.org/tblock/tblock/pulls/17
if os.name == 'nt':
    # On POSIX, this breaks the colors
    init(convert=True)


class Font:
    BOLD = '\033[1m'
    DEFAULT = '\033[0m'
    UNDERLINE = '\033[4m'

    # The check mark appears weird on Windows, so replace it by 'v'
    if os.name == 'nt':
        CHECK_MARK = 'v'
    else:
        CHECK_MARK = '✓'


class Var:

    # The default IP address where to redirect blocked domains
    DEFAULT_IP = '127.0.0.1'

    # The list of all official mirrors of the remote filter repository that are available
    # Read more: https://tblock.codeberg.page/wiki/filters/remote-repository#mirrors
    REPO_MIRRORS = [
        'https://codeberg.org/tblock/remote-filters-repository/raw/branch/main/index.xml',
        'https://0xacab.org/twann/remote-filters-repository-mirror/-/raw/main/index.xml',
        'https://git.disroot.org/tblock/remote-filters-repository/raw/branch/main/index.xml',
        'https://tblock.codeberg.page/repo/index.xml'
    ]

    # Shuffle the mirror list to avoid too much traffic on only one of them
    random.shuffle(REPO_MIRRORS)


class Path:

    # The script is running on Termux
    if os.path.isdir('/data/data/com.termux/files/usr/lib/'):
        PREFIX = '/data/data/com.termux/files/usr/lib/tblock/'
        HOSTS = '/system/etc/hosts'
        TMP_DIR = '/data/data/com.termux/files/usr/tmp/tblock/'

    # The script is running on a POSIX platform
    elif os.name == 'posix':
        PREFIX = '/var/lib/tblock/'
        HOSTS = '/etc/hosts'
        TMP_DIR = '/tmp/tblock/'

    # The script is running on Windows
    elif os.name == 'nt':
        PREFIX = os.path.join(os.path.expandvars('%ALLUSERSPROFILE%'), 'TBlock')
        HOSTS = os.path.join(os.path.expandvars('%WINDIR%'), 'System32', 'drivers', 'etc', 'hosts')
        TMP_DIR = os.path.join(os.path.expandvars('%TMP%'), 'tblock')

    # If the script is running on an unsupported platform, raise an error
    else:
        raise OSError('TBlock is currently not supported on your operating system')

    # Other paths
    DATABASE = os.path.join(PREFIX, 'storage.sqlite')
    HOSTS_BACKUP = os.path.join(PREFIX, 'hosts')
    REPO_VERSION = os.path.join(PREFIX, 'repo')
    HOSTS_VERIFICATION = os.path.join(PREFIX, 'hosts_protection')
    DB_LOCK = os.path.join(PREFIX, '.db-lock')
    NEEDS_UPDATE = os.path.join(PREFIX, '.needs-update')


def setup_database(db_path: str) -> None:
    """Setup the SQLite3 database that is needed by TBlock in order to work
    Args:
        db_path (str): The path to the database to setup
    """
    with sqlite3.connect(db_path) as db:
        db.cursor().execute('''CREATE TABLE IF NOT EXISTS "rules" (
            "domain"	TEXT NOT NULL UNIQUE,
            "policy"	TEXT NOT NULL,
            "filter_id"	TEXT NOT NULL,
            "priority"	INTEGER NOT NULL,
            "ip"	TEXT,
            PRIMARY KEY("domain")
        );''')
        db.cursor().execute('''CREATE TABLE IF NOT EXISTS "filters" (
            "id"	TEXT NOT NULL UNIQUE,
            "source"	TEXT NOT NULL UNIQUE,
            "metadata"	TEXT NOT NULL,
            "subscribing"	INTEGER NOT NULL,
            "on_rfr"	INTEGER NOT NULL,
            "permissions"   TEXT,
            PRIMARY KEY("id")
        );''')
        db.commit()


# NOTE: These lines should be executed every time TBlock is launched

def check_dirs() -> None:
    """Check and create directories and files that doesn't exist
    """
    # Check if the hosts file exists. If it doesn't, raise an error
    if not os.path.isfile(Path.HOSTS):
        raise OSError('TBlock is currently not supported on your operating system')

    # Create directories if they do not exist
    for d in [Path.PREFIX, Path.TMP_DIR]:
        if not os.path.isdir(d):
            try:
                os.mkdir(d)
            except PermissionError:
                continue

    # Setup the database if it does not exist
    if not os.path.isfile(Path.DATABASE):
        try:
            setup_database(Path.DATABASE)
        except (PermissionError, sqlite3.OperationalError):
            pass


# Check if TBlock is protecting the machine or not
global PROTECTING_STATUS
if os.path.isfile(Path.HOSTS_BACKUP):
    PROTECTING_STATUS = True
else:
    PROTECTING_STATUS = False



#!/usr/bin/env python

import os
import tblock.config
from tblock.tools import enable_protection, disable_protection
import unittest

tblock.config.Path.HOSTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts')
tblock.config.Path.HOSTS_BACKUP = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts.bak')
tblock.config.Path.DATABASE = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'storage.sqlite')
tblock.config.Path.DB_LOCK = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', '.db-lock')
tblock.config.Path.TMP_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root')
tblock.config.Path.HOSTS_VERIFICATION = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'verification')
tblock.config.setup_database(tblock.config.Path.DATABASE)


class TestHostsFile(unittest.TestCase):

    def test_restore_hosts(self):
        self.assertEquals(
            disable_protection(True),
            None
        )

    def test_update_hosts(self):
        self.assertEquals(
            enable_protection(True),
            None
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


# Standard libraries
import sqlite3
import os.path
import json
import re

# External libraries
from colorama import Fore, Style
import requests
import urllib3

# Local libraries
from .rules import get_all_rules
from .converter.parser import FilterParser
from .config import Path, Font
from .utils import check_root, prompt_user, is_url, download_file, lock_database, unlock_database
from .hosts import update_hosts
from .exceptions import FilterExistsError, FilterSourceExistsError, AlreadySubscribingError, InvalidLinkError, \
    FilterNotExists, NotSubscribingError, SIGTERM, NetworkError, MissingArgumentError, FilterReservedError, \
    InvalidFilterSyntax, FilterNotCustomError


class FilterPermissions:

    def __init__(self, pattern: str = None) -> None:
        """Filter permission object (which rule policies a filter is allowed to use)

        Args:
            pattern (str, optional): The pattern that contains permissions
        """

        # By default, is a permission pattern is empty, the only policy allowed will be blocked
        # Therefore, it is impossible to restrict a filter to add all types of rule policies. The only way to do that is
        # to unsubscribe to the filter. There would be no reason to keep a filter as 'subscribed' if that filter was
        # not allowed to set any rules; it would simply slow down TBlock, since it would still need to download the
        # source of the filter.
        if not pattern:
            self.block = True
            self.allow = False
            self.redirect = False

        else:
            if re.match(re.compile(r'.*b', re.IGNORECASE), pattern):
                self.block = True
            else:
                self.block = False
            if re.match(re.compile(r'.*a', re.IGNORECASE), pattern):
                self.allow = True
            else:
                self.allow = False
            if re.match(re.compile(r'.*r', re.IGNORECASE), pattern):
                self.redirect = True
            else:
                self.redirect = False

        # Cleanly redefine permission pattern to avoid multiple 'a', 'b' or 'r', or uppercase letters
        self.pattern = ''
        self.pattern += 'a' if self.allow else ''
        self.pattern += 'b' if self.block else ''
        self.pattern += 'r' if self.redirect else ''


class Filter:

    def __init__(self, filter_id: str) -> None:
        """Filter object

        Args:
            filter_id (str): The ID of the filter
        """

        if filter_id == '!user':
            raise FilterReservedError('the filter ID "!user" is reserved by TBlock and cannot be used')
        self.id = filter_id

        # Fetch filter data from the database
        with sqlite3.connect(Path.DATABASE) as conn:
            data = conn.cursor().execute(
                'SELECT "source", "metadata", "subscribing", "on_rfr", "permissions" FROM "filters" WHERE id=?;',
                (self.id,)
            ).fetchall()
        self.exists = bool(data)
        self.source = data[0][0] if data else None
        self.metadata = json.loads(data[0][1]) if data else {}
        self.subscribing = bool(data[0][2]) if data else False
        self.on_rfr = bool(data[0][3]) if data else False
        self.permissions = FilterPermissions(data[0][4]) if data else None

    def subscribe(self, permissions: FilterPermissions, source: str = None) -> None:
        """Mark a filter as subscribed in the database

        Args:
            permissions (FilterPermissions): Permissions of the filter
            source (str, optional): If the filter is a custom filter, specify its source
        """
        self.permissions = permissions
        print(f'{Font.BOLD}==> Subscribing to filter: {self.id}{Font.DEFAULT}')

        if source:
            if self.exists:
                raise FilterExistsError(f'the filter "{self.id}" already exists')
            else:
                if source in get_all_sources():
                    raise FilterSourceExistsError(f'the source "{source}" is already the source of another filter')
                else:
                    self.source = source
                    print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Marking filter as subscribed in database', end='\r')
                    with sqlite3.connect(Path.DATABASE) as conn:
                        conn.cursor().execute(
                            'INSERT INTO "filters" ("id", "source", "metadata", "subscribing", "on_rfr", "permissions")'
                            'VALUES (?, ?, ?, ?, ?, ?);',
                            (self.id, self.source, json.dumps(self.metadata), True, False, self.permissions.pattern)
                        )
                        conn.commit()
                    self.exists = True
                    self.subscribing = True
                    print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Marking filter as subscribed in database')

        else:
            if not self.exists:
                raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')
            else:
                if self.subscribing:
                    raise AlreadySubscribingError(f'you are already subscribing to filter: "{self.id}"')
                else:
                    print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Marking filter as subscribed in database', end='\r')
                    with sqlite3.connect(Path.DATABASE) as conn:
                        conn.cursor().execute(
                            'UPDATE "filters" SET subscribing=?, permissions=? WHERE id=?;',
                            (True, permissions.pattern, self.id)
                        )
                        conn.commit()
                    self.subscribing = True
                    print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Marking filter as subscribed in database')

    def change_permissions(self, permissions: FilterPermissions) -> None:
        """Change the permissions of a filter

        Args:
            permissions (FilterPermissions): New permissions of the filter
        """
        if not self.exists:
            raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')
        elif not self.subscribing:
            raise NotSubscribingError(f'you are not subscribing to filter "{self.id}"')
        print(f'{Font.BOLD}==> Changing permissions of filter: {self.id}{Font.DEFAULT}')
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Writing new permissions in database', end='\r')
        with sqlite3.connect(Path.DATABASE) as conn:
            conn.cursor().execute(
                'UPDATE "filters" SET permissions=? WHERE id=?;',
                (permissions.pattern, self.id)
            )
            conn.commit()
        self.permissions = permissions
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Writing new permissions in database')

    def unsubscribe(self) -> None:
        """Unsubscribe from a filter. If the filter is available on the remote filter repository, it will simply be
           marked as "unsubscribed", but if it is a custom filter, it will be removed from the database.
        """
        if not self.exists:
            raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')
        elif not self.subscribing:
            raise NotSubscribingError(f'you are not subscribing to filter "{self.id}"')
        print(f'{Font.BOLD}==> Unsubscribing from filter: {self.id}{Font.DEFAULT}')
        with sqlite3.connect(Path.DATABASE) as conn:

            # Simply mark as "unsubscribed" if filter is available on the remote repository
            if self.on_rfr:
                print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Marking filter as subscribed in database', end='\r')
                conn.cursor().execute(
                    'UPDATE "filters" SET subscribing=?, permissions=? WHERE id=?;',
                    (False, None, self.id)
                )
                self.subscribing = False
                print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Marking filter as subscribed in database')

            # Remove filter from the database if it is a custom filter
            else:
                print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Removing custom filter from database', end='\r')
                conn.cursor().execute(
                    'DELETE FROM "filters" WHERE id=?;',
                    (self.id,)
                )
                self.exists = False
                self.subscribing = False
                print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Removing custom filter from database')
            conn.commit()

    def delete_all_rules(self) -> None:
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Removing all filter rules from database', end='\r')
        with sqlite3.connect(Path.DATABASE) as conn:
            conn.cursor().execute(
                'DELETE FROM "rules" WHERE filter_id=?;',
                (self.id,)
            )
            conn.commit()
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Removing all filter rules from database')

    def change_id(self, new_id: str) -> None:
        """Change the ID of a custom filter

        Args:
            new_id (str): The new filter ID to use
        """
        if not self.exists:
            raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')
        elif self.on_rfr:
            raise FilterNotCustomError(f'the filter "{self.id}" is not a custom filter')
        print(f'{Font.BOLD}==> Changing filter ID: {self.id}{Font.DEFAULT}')
        self.delete_all_rules()
        with sqlite3.connect(Path.DATABASE) as conn:
            conn.cursor().execute(
                'UPDATE "filters" SET id=? WHERE id=?;',
                (new_id, self.id)
            )
        self.update(False)

    def update(self, verbosity: bool = True) -> None:
        """Update rules from a filter

        Args:
            verbosity (bool, optional): Display info about what is happening (default)
        """
        if not self.exists:
            raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')
        elif not self.subscribing:
            raise NotSubscribingError(f'you are not subscribing to filter "{self.id}"')
        if verbosity:
            print(f'{Font.BOLD}==> Updating filter: {self.id}{Font.DEFAULT}')

        # Download the filter is it is online
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Get: {self.source}', end='\r')
        if is_url(self.source):
            tmp_filter = os.path.join(Path.TMP_DIR, self.id + '.tmp')
            try:
                download_file(self.source, tmp_filter)
            except (requests.exceptions.ConnectionError, urllib3.exceptions.MaxRetryError, ConnectionRefusedError,
                    urllib3.exceptions.NewConnectionError, InvalidLinkError):
                print(f' [{Fore.RED}x{Style.RESET_ALL}] Get: {self.source}')
                raise NetworkError(f'{Fore.RED}ERROR: cannot download: {self.source}{Style.RESET_ALL}')
        else:
            # This sequence is just meant to show an error message in the terminal before exiting
            tmp_filter = self.source
            try:
                with open(tmp_filter, 'rt') as f:
                    f.close()
            except FileNotFoundError:
                print(f' [{Fore.RED}x{Style.RESET_ALL}] Get: {self.source}')
                raise FileNotFoundError(f'cannot find file: {self.source}')
            else:
                print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Get: {self.source}', end='\r')
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Get: {self.source}')

        # Check if the filter has a specified syntax
        try:
            rules = FilterParser(tmp_filter, syntax=self.metadata['syntax'])
        except KeyError:
            rules = FilterParser(tmp_filter)

        self.delete_all_rules()

        # Get all rules from the filter and from the database
        all_rules = rules.get_filter_content(
            self.permissions.allow, self.permissions.block, self.permissions.redirect, False
        )
        all_rules_in_db = get_all_rules()

        # Remove the filter only if it has been downloaded in a temporary location
        if self.source != tmp_filter:
            os.remove(tmp_filter)

        count = 1
        all_rules_count = len(all_rules)
        db = sqlite3.connect(Path.DATABASE)
        cursor = db.cursor()
        load_character = '|'

        # Start to add rules in the database
        for rule in all_rules:

            if count < all_rules_count:
                print(f' [{Fore.YELLOW}{load_character}{Style.RESET_ALL}] Adding new rules to database '
                      f'({count}/{all_rules_count})', end='\r')
                # Show process
                if load_character == '|' and not count % 500:
                    load_character = '/'
                elif load_character == '/' and not count % 500:
                    load_character = '-'
                elif load_character == '-' and not count % 500:
                    load_character = '\\'
                elif load_character == '\\' and not count % 500:
                    load_character = '|'
            else:
                print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Adding new rules to database '
                      f'({count}/{all_rules_count})')

            if rule[0] == 'rule':

                if rule[1] in all_rules_in_db.keys() and all_rules_in_db[rule[1]]['priority']:
                    continue

                elif rule[2] == 'allow' and self.permissions.allow or rule[2] == 'block' and self.permissions.block:

                    # Avoid allowing or redirecting rules to be overwritten by blocking rules
                    # See: https://codeberg.org/tblock/tblock/issues/4
                    if rule[1] in all_rules_in_db.keys():
                        if rule[2] == 'block' and all_rules_in_db[rule[1]]['policy'] in ['redirect', 'allow']:
                            continue
                        cursor.execute(
                            'UPDATE "rules" SET policy=?, filter_id=?, priority=? WHERE domain=?;',
                            (rule[2], self.id, False, rule[1])
                        )
                    else:
                        cursor.execute(
                            'INSERT INTO "rules" ("domain", "policy", "filter_id", "priority") VALUES (?, ?, ?, ?)',
                            (rule[1], rule[2], self.id, False)
                        )
                        all_rules_in_db[rule[1]] = {
                            'policy': rule[2],
                            'filter_id': self.id,
                            'priority': False,
                            'ip': None
                        }

                elif rule[2] == 'redirect' and self.permissions.redirect:

                    # Avoid allowing rules to be overwritten by redirecting rules
                    # See: https://codeberg.org/tblock/tblock/issues/4
                    if rule[1] in all_rules_in_db.keys():
                        if all_rules_in_db[rule[1]]['policy'] == 'allow':
                            continue
                        cursor.execute(
                            'UPDATE "rules" SET policy=?, filter_id=?, priority=?, ip=? WHERE domain=?;',
                            (rule[2], self.id, False, rule[3], rule[1])
                        )
                    else:
                        cursor.execute(
                            'INSERT INTO "rules" ("domain", "policy", "filter_id", "priority", "ip") '
                            'VALUES (?, ?, ?, ?, ?)',
                            (rule[1], rule[2], self.id, False, rule[3])
                        )
                        all_rules_in_db[rule[1]] = {
                            'policy': rule[2],
                            'filter_id': self.id,
                            'priority': False,
                            'ip': rule[3]
                        }
            else:
                continue

            count += 1
        db.commit()
        db.close()
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Filter successfully updated')

    def show(self) -> None:
        """Show information about a filter
        """
        if not self.exists:
            raise FilterNotExists(f'the filter "{self.id}" does not exist or is not available')

        # If subscribing to filter, get the number of rules that are set by it
        if self.subscribing:
            with sqlite3.connect(Path.DATABASE) as db:
                total_rules = len(
                    db.cursor().execute(
                        'SELECT "domain" FROM "rules" WHERE filter_id=?;',
                        (self.id,)
                    ).fetchall()
                )
        else:
            total_rules = None

        print(f'Filter ID     : {self.id}')

        if self.on_rfr:
            print(f'Title         : {self.metadata["title"]}')
            print(f'Homepage      : {self.metadata["homepage"]}')
            print(f'License       : {self.metadata["license"]}')
            print(f'Syntax        : {self.metadata["syntax"]}')
            print('Type          : available on online repository')

        else:
            print('Type          : custom filter')

        print(f'Source        : {self.source}')

        if self.subscribing:
            print('Subscribing   : yes')
            print(f'Permissions   : {self.permissions.pattern}')
            print(f'Total rules   : {total_rules}')
        else:
            print('Subscribing   : no')

        # This must be the last thing displayed, as it can be long and break the visual rendering of other information
        if self.on_rfr:
            print(f'Description   : {self.metadata["description"]}')


def get_all_sources() -> list:
    """Get a list of all filter sources from the database

    Returns:
        list: A list of all sources
    """
    with sqlite3.connect(Path.DATABASE) as conn:
        data = conn.execute('SELECT "source" FROM "filters";').fetchall()
    for row in data:
        data[data.index(row)] = row[0]
    return data


def get_all_custom() -> list:
    """Get a list of all custom filter IDs

    Returns:
        list: A list of all IDs
    """
    with sqlite3.connect(Path.DATABASE) as conn:
        data = conn.execute('SELECT "id" FROM "filters" WHERE on_rfr=0;').fetchall()
    for row in data:
        data[data.index(row)] = row[0]
    return data


def get_all_id(subscribing: bool = True, not_subscribing: bool = True) -> list:
    """Get a list of all filter IDs from the database

    Args:
        subscribing (bool, optional): Also return subscribed filters (default)
        not_subscribing (bool, optional): Also return not subscribed filters (default)

    Returns:
        list: A list of all IDs
    """
    with sqlite3.connect(Path.DATABASE) as conn:
        if subscribing and not_subscribing:
            data = conn.execute('SELECT "id" FROM "filters";').fetchall()
        elif subscribing:
            data = conn.execute('SELECT "id" FROM "filters" WHERE subscribing=1;').fetchall()
        elif not_subscribing:
            data = conn.execute('SELECT "id" FROM "filters" WHERE subscribing=0;').fetchall()
        else:
            data = []
    for row in data:
        data[data.index(row)] = row[0]
    return data


def get_all_from_rfr(subscribing: bool = True, not_subscribing: bool = True) -> list:
    """Get a list of all filter IDs available on the remote repository

    Returns:
        list: A list of all IDs
    """
    with sqlite3.connect(Path.DATABASE) as conn:
        if subscribing and not_subscribing:
            data = conn.execute('SELECT "id" FROM "filters" WHERE on_rfr=1;').fetchall()
        elif subscribing:
            data = conn.execute('SELECT "id" FROM "filters" WHERE on_rfr=1 AND subscribing=1;').fetchall()
        elif not_subscribing:
            data = conn.execute('SELECT "id" FROM "filters" WHERE on_rfr=1 AND subscribing=0;').fetchall()
        else:
            data = []
    for row in data:
        data[data.index(row)] = row[0]
    return data


def subscribe_to_filters(filters: list, force: bool = False) -> None:
    """Subscribe to a list of filters

    Args:
        filters (list): All filters ID in a list: [ [id, permissions, custom_source (if any)] ]
        force (bool, optional): Do not prompt user before subscribing to filters (False by default)
    """
    list_id = []
    for filter_id in filters:
        if not Filter(filter_id[0]).exists and len(filter_id) < 3:
            raise FilterNotExists(f'{Fore.RED}ERROR: filter does not exist: {filter_id[0]}{Style.RESET_ALL}')
        list_id.append(filter_id[0])
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not filters:
        raise MissingArgumentError(f'{Fore.RED}ERROR: you need to specify at least one filter{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to subscribe to the following filters:', list_id):
        raise SIGTERM()
    print(':: Subscribing to filters...')
    lock_database()
    try:
        for filter_data in filters:
            filter_obj = Filter(filter_data[0])
            try:
                if len(filter_data) == 1:
                    filter_obj.subscribe(FilterPermissions())
                elif len(filter_data) == 2:
                    filter_obj.subscribe(FilterPermissions(filter_data[1]))
                else:
                    filter_obj.subscribe(FilterPermissions(filter_data[1]), filter_data[2])
            except AlreadySubscribingError as err:
                print(f'{Fore.YELLOW}WARNING: {err}{Style.RESET_ALL}')
            except (FilterSourceExistsError, FilterExistsError) as err:
                raise FilterSourceExistsError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
            try:
                filter_obj.update(False)
            except InvalidFilterSyntax as err:
                raise InvalidFilterSyntax(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
            except FileNotFoundError as err:
                raise FileNotFoundError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
        update_hosts(get_all_rules())
        unlock_database()
    except PermissionError:
        raise PermissionError(f'{Fore.RED}ERROR: permission error{Style.RESET_ALL}')


def change_filters_permissions(filters: list, force: bool = False) -> None:
    """Subscribe to a list of filters

    Args:
        filters (list): All filters ID in a list: [ [id, permissions] ]
        force (bool, optional): Do not prompt user (False by default)
    """
    list_id = []
    for filter_id in filters:
        if not Filter(filter_id[0]).subscribing:
            raise NotSubscribingError(f'{Fore.RED}ERROR: not subscribing to filter: {filter_id[0]}{Style.RESET_ALL}')
        list_id.append(filter_id[0])
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not filters:
        raise MissingArgumentError(f'{Fore.RED}ERROR: you need to specify at least one filter{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to change the permissions of the following filters:', list_id):
        raise SIGTERM()
    print(':: Changing filters permissions...')
    lock_database()
    try:
        for filter_data in filters:
            filter_obj = Filter(filter_data[0])
            try:
                filter_obj.change_permissions(FilterPermissions(filter_data[1]))
            except NotSubscribingError as err:
                print(f'{Fore.YELLOW}WARNING: {err}{Style.RESET_ALL}')
            try:
                filter_obj.update(False)
            except InvalidFilterSyntax as err:
                raise InvalidFilterSyntax(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
            except FileNotFoundError as err:
                raise FileNotFoundError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
        update_hosts(get_all_rules())
        unlock_database()
    except PermissionError:
        raise PermissionError(f'{Fore.RED}ERROR: permission error{Style.RESET_ALL}')


def unsubscribe_from_filters(filters: list, force: bool = False) -> None:
    """Subscribe from a list of filters

    Args:
        filters (list): All filters ID in a list: [id_1, id_2]
        force (bool, optional): Do not prompt user (False by default)
    """
    list_id = []
    for filter_id in filters:
        if not Filter(filter_id).subscribing:
            raise NotSubscribingError(f'{Fore.RED}ERROR: not subscribing to filter: {filter_id}{Style.RESET_ALL}')
        list_id.append(filter_id)
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not filters:
        raise MissingArgumentError(f'{Fore.RED}ERROR: you need to specify at least one filter{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to unsubscribe from the following filters:', list_id):
        raise SIGTERM()
    print(':: Unsubscribing from filters...')
    lock_database()
    try:
        for filter_id in filters:
            filter_obj = Filter(filter_id)
            filter_obj.unsubscribe()
            filter_obj.delete_all_rules()
        update_hosts(get_all_rules())
        unlock_database()
    except PermissionError:
        raise PermissionError(f'{Fore.RED}ERROR: permission error{Style.RESET_ALL}')
    try:
        open(Path.NEEDS_UPDATE, 'wt').close()
    except PermissionError:
        pass


def update_filters(filters: list, force: bool = False) -> None:
    """Update rules from a list of filters

    Args:
        filters (list): All filters ID in a list: [id_1, id_2]
        force (bool, optional): Do not prompt user (False by default)
    """
    list_id = []
    for filter_id in filters:
        if not Filter(filter_id).subscribing:
            raise NotSubscribingError(f'{Fore.RED}ERROR: not subscribing to filter: {filter_id}{Style.RESET_ALL}')
        list_id.append(filter_id)
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not filters:
        MissingArgumentError(f'{Fore.RED}ERROR: you need to specify at least one filter{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to update the rules from the following filters:', list_id):
        raise SIGTERM()
    print(':: Updating filters...')
    lock_database()
    try:
        for filter_id in filters:
            filter_obj = Filter(filter_id)
            try:
                filter_obj.update()
            except InvalidFilterSyntax as err:
                raise InvalidFilterSyntax(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
            except FileNotFoundError as err:
                raise FileNotFoundError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
        update_hosts(get_all_rules())
        unlock_database()
    except PermissionError:
        raise PermissionError(f'{Fore.RED}ERROR: permission error{Style.RESET_ALL}')


def change_filter_id(filter_id: str, new_id: str, force: bool = False) -> None:
    """Change the ID of a custom filter

    Args:
        filter_id (str): The filter to rename
        new_id (str): The new ID to use
        force (bool, optional): Do not prompt user (False by default)
    """
    if not Filter(filter_id).subscribing:
        raise NotSubscribingError(f'{Fore.RED}ERROR: not subscribing to filter: {filter_id}{Style.RESET_ALL}')
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    if Filter(new_id).exists:
        raise FilterExistsError(
            f'{Fore.RED}ERROR: the filter ID "{new_id}" is already taken by another filter{Style.RESET_ALL}'
        )
    elif not force and not prompt_user(f'You are about to change the ID of the filter "{filter_id}" to:', [new_id]):
        raise SIGTERM()
    print(':: Changing filter ID...')
    lock_database()
    try:
        filter_obj = Filter(filter_id)
        filter_obj.change_id(new_id)
        update_hosts(get_all_rules())
        unlock_database()
    except FilterNotCustomError as err:
        raise FilterNotCustomError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    except PermissionError:
        raise PermissionError(f'{Fore.RED}ERROR: permission error{Style.RESET_ALL}')


def update_all_filters(force: bool = False) -> None:
    """Update rules from all filters that the user is subscribing to

    Args:
        force (bool, optional): Do not prompt user (False by default)
    """
    update_filters(get_all_id(subscribing=True, not_subscribing=False), force)
    try:
        os.remove(Path.NEEDS_UPDATE)
    except (FileNotFoundError, PermissionError):
        pass


def show_filters(filters: list) -> None:
    """Show information about a list of filters

    Args:
        filters (list): All filters ID in a list: [id_1, id_2]
    """
    if not filters:
        raise MissingArgumentError(f'{Fore.RED}ERROR: please specify at least one filter{Style.RESET_ALL}')
    for filter_id in filters:
        if not Filter(filter_id).exists:
            raise FilterNotExists(f'{Fore.RED}ERROR: filter does not exist: {filter_id}{Style.RESET_ALL}')
    for filter_id in filters:
        filter_obj = Filter(filter_id)
        print('----------------------------------------------')
        filter_obj.show()
    print('----------------------------------------------')


def list_filters(online_only: bool = False, custom_only: bool = False,
                 subscribing_only: bool = True, not_subscribing_only: bool = False) -> None:
    """List all filters with some criteria

    Args:
        online_only (bool, optional): List only filters that are available on the remote repository
        custom_only (bool, optional): List only custom filters
        subscribing_only (bool, optional): List only filters that the user is subscribing to
        not_subscribing_only (bool, optional): List only filters that the user is not subscribing to
    """

    if subscribing_only and not_subscribing_only:
        subscribing = False
        not_subscribing = False
    elif subscribing_only and not not_subscribing_only:
        subscribing = True
        not_subscribing = False
    elif not_subscribing_only and not subscribing_only:
        subscribing = False
        not_subscribing = True
    else:
        subscribing = True
        not_subscribing = True

    if online_only and custom_only:
        pass
    elif online_only and not custom_only:
        for filter_id in get_all_from_rfr(subscribing, not_subscribing):
            print(filter_id)
    elif custom_only and not online_only:
        for filter_id in get_all_custom():
            print(filter_id)
    else:
        for filter_id in get_all_id(subscribing, not_subscribing):
            print(filter_id)


def search_filters(query: str, online_only: bool = False, custom_only: bool = False,
                   subscribing_only: bool = False, not_subscribing_only: bool = False) -> None:
    """Search a query in the filters list and their metadata

    Args:
        query (str): The query to search
        online_only (bool, optional): Show only results from filters that are available on the remote repository
        custom_only (bool, optional): Show only results from custom filters
        subscribing_only (bool, optional): Show only results from filters that the user is subscribing to
        not_subscribing_only (bool, optional): Show only results from filters that the user is not subscribing to
    """
    regex_query = re.compile(query, re.IGNORECASE)
    if subscribing_only and not_subscribing_only:
        subscribing = False
        not_subscribing = False
    elif subscribing_only and not not_subscribing_only:
        subscribing = True
        not_subscribing = False
    elif not_subscribing_only and not subscribing_only:
        subscribing = False
        not_subscribing = True
    else:
        subscribing = True
        not_subscribing = True

    if online_only and custom_only:
        filters_list = []
    elif online_only and not custom_only:
        filters_list = get_all_from_rfr(subscribing, not_subscribing)
    elif custom_only and not online_only:
        filters_list = get_all_custom()
    else:
        filters_list = get_all_id(subscribing, not_subscribing)

    for filter_id in filters_list:
        if re.findall(regex_query, Filter(filter_id).id):
            print(filter_id)
        elif re.findall(regex_query, Filter(filter_id).metadata['title']):
            print(filter_id)
        elif re.findall(regex_query, Filter(filter_id).metadata['description']):
            print(filter_id)
        else:
            pass



#!/usr/bin/env python

import os
import tblock.config
import tblock.exceptions
from tblock import filters
from nose.tools import raises
import unittest

tblock.config.Path.HOSTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts')
tblock.config.Path.DATABASE = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'storage.sqlite')
tblock.config.Path.DB_LOCK = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', '.db-lock')
tblock.config.Path.TMP_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root')
tblock.config.Path.HOSTS_VERIFICATION = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'verification')
tblock.config.setup_database(tblock.config.Path.DATABASE)

test_custom_filter = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'filter.txt')


class TestSubscribeFilters(unittest.TestCase):

    def test_subscribe_to_custom(self):
        self.assertEquals(
            filters.subscribe_to_filters([['test-filter', 'b', test_custom_filter]], force=True),
            None
        )

    def test_subscribe_to_rfr(self):
        self.assertEquals(
            filters.subscribe_to_filters([['twann-list', 'b'], ['twann-anti-ip-loggers', 'b']], force=True),
            None
        )

    @raises(tblock.exceptions.InvalidFilterSyntax)
    def test_subscribe_to_unsupported_filetype(self):
        self.assertEquals(
            filters.subscribe_to_filters([['test-filter-error', 'b', tblock.config.Path.DATABASE]], force=True),
            None
        )

    def test_subscribe_to_unsupported_filetype_remove(self):
        try:
            os.remove(tblock.config.Path.DB_LOCK)
        except FileNotFoundError:
            pass
        self.assertEquals(
            filters.unsubscribe_from_filters(['test-filter-error'], force=True),
            None
        )


class TestSubscribingChangeID(unittest.TestCase):

    def test_change_filter_id(self):
        self.assertEquals(
            filters.change_filter_id('test-filter', 'test-filter2', force=True),
            None
        )


class TestSubscribingPermissions(unittest.TestCase):

    def test_change_filters_permissions(self):
        self.assertEquals(
            filters.change_filters_permissions([['twann-anti-ip-loggers', 'ar']], force=True),
            None
        )


class TestSubscribingUpdate(unittest.TestCase):

    def test_update_filter(self):
        self.assertEquals(
            filters.update_filters(['twann-list'], force=True),
            None
        )

    def test_update_filters_all(self):
        self.assertEquals(
            filters.update_all_filters(force=True),
            None
        )


class TestUnsubscribeFilters(unittest.TestCase):

    def test_unsubscribe_from_custom(self):
        self.assertEquals(
            filters.unsubscribe_from_filters(['test-filter2'], force=True),
            None
        )

    def test_unsubscribe_from_rfr(self):
        self.assertEquals(
            filters.unsubscribe_from_filters(['twann-list', 'twann-anti-ip-loggers'], force=True),
            None
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import sqlite3

# External libraries
from colorama import Fore, Style

# Local libraries
from .hosts import update_hosts
from .config import Path, Font
from .utils import is_valid_ip, extract_domain, check_root, prompt_user, lock_database, unlock_database
from .exceptions import InvalidBindAddress, RuleExistsError, InvalidRulePolicy, \
    RuleNotExistsError, MissingArgumentError, SIGTERM


class Rule:

    def __init__(self, domain: str) -> None:
        """Rule object to use for user rules. This is to slow to be used for filter rules

        Args:
            domain (str): The domain or the URL of the domain to manage
        """
        self.domain = extract_domain(domain)

        # Fetch rule data from the database
        with sqlite3.connect(Path.DATABASE) as conn:
            data = conn.cursor().execute(
                'SELECT "policy", "priority", "filter_id", "ip" FROM "rules" WHERE domain=?;',
                (self.domain,)
            ).fetchall()
        self.exists = bool(data)
        self.policy = data[0][0] if data else None
        self.priority = data[0][1] if data else False
        self.filter_id = data[0][2] if data else None
        self.ip = data[0][3] if data else None

    def add(self, policy: str, ip: str = None) -> None:
        """Add rule to database

        Args:
            policy (str): The policy of the rule (allow/block/redirect)
            ip (str, optional): The IP where to redirect domain
        """
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Adding rule for {self.domain}', end='\r')

        if policy not in ['allow', 'block', 'redirect']:
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Adding rule for {self.domain}')
            raise InvalidRulePolicy(f'invalid rule policy: "{policy}"')
        elif policy == 'redirect' and not ip:
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Adding rule for {self.domain}')
            raise InvalidBindAddress('redirecting IP address cannot be NoneType')
        elif policy == 'redirect' and not is_valid_ip(ip):
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Adding rule for {self.domain}')
            raise InvalidBindAddress('redirecting IP address cannot be NoneType')
        elif policy == self.policy and ip == self.ip and self.priority:
            print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Adding rule for {self.domain}')
            raise RuleExistsError(f'exact same rule already exists for domain: "{self.domain}"')

        with sqlite3.connect(Path.DATABASE) as db:
            if self.exists:
                db.cursor().execute(
                    'UPDATE "rules" SET policy=?, priority=?, filter_id=?, ip=?  WHERE domain=?;',
                    (policy, True, '!user', ip, self.domain)
                )
            else:
                db.cursor().execute(
                    'INSERT INTO "rules" ("domain", "policy", "priority", "filter_id", "ip") VALUES (?, ?, ?, ?, ?);',
                    (self.domain, policy, True, '!user', ip)
                )
            db.commit()
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Adding rule for {self.domain}')

    def delete(self) -> None:
        """Delete an user rule from database
        """
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Deleting rule for {self.domain}', end='\r')

        if not self.priority:
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Deleting rule for {self.domain}')
            raise RuleNotExistsError(f'no rule found for domain "{self.domain}"')

        with sqlite3.connect(Path.DATABASE) as db:
            db.cursor().execute(
                'DELETE FROM "rules" WHERE domain=?;',
                (self.domain,)
            )
            db.commit()
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Deleting rule for {self.domain}')


def get_all_rules(standard: bool = True, user: bool = True) -> dict:
    """Retrieve all active rules from the database

    Args:
        standard (bool, optional): Include rules set by filters
        user (bool, optional): Include user rules

    Returns:
        dict: All rules in the directory. The domain is the directory key
    """
    all_rules = {}
    with sqlite3.connect(Path.DATABASE) as db:
        if standard and user:
            data = db.cursor().execute(
                'SELECT "domain","policy","filter_id","priority","ip" FROM "rules";'
            ).fetchall()
        elif standard:
            data = db.cursor().execute(
                'SELECT "domain","policy","filter_id","priority","ip" FROM "rules" WHERE priority=0;'
            ).fetchall()
        elif user:
            data = db.cursor().execute(
                'SELECT "domain","policy","filter_id","priority","ip" FROM "rules" WHERE priority=1;'
            ).fetchall()
        else:
            data = []
    for row in data:
        all_rules[row[0]] = {}
        all_rules[row[0]]['policy'] = row[1]
        all_rules[row[0]]['filter_id'] = row[2]
        all_rules[row[0]]['priority'] = row[3]
        all_rules[row[0]]['ip'] = row[4]
    return all_rules


def add_rules(rules: list, force: bool = False) -> None:
    """Add rules for domains

    Args:
        rules (list): All rules in a list: [ [domain, policy, ip (if any)] ]
        force (bool, optional): Do not prompt user (False by default)
    """
    list_domains = []
    for r in rules:
        list_domains.append(extract_domain(r[0]))
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not rules:
        raise MissingArgumentError(
            f'{Fore.RED}ERROR: you need to specify at least one domain to manage{Style.RESET_ALL}'
        )
    elif not force and not prompt_user('You are about to add rules for the following domains:', list_domains):
        raise SIGTERM()
    print(':: Adding rules...')
    lock_database()
    for rule_data in rules:
        rule_obj = Rule(rule_data[0])
        try:
            if len(rule_data) == 2:
                rule_obj.add(rule_data[1])
            else:
                rule_obj.add(rule_data[1], rule_data[2])
        except RuleExistsError as err:
            print(f'{Fore.YELLOW}WARNING: {err}{Style.RESET_ALL}')
        except InvalidBindAddress as err:
            raise InvalidBindAddress(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    update_hosts(get_all_rules())
    unlock_database()


def delete_rules(rules: list, force: bool = False) -> None:
    """Delete rules for domains

    Args:
        rules (list): All rules in a list: [domain1, domain2]
        force (bool, optional): Do not prompt user (False by default)
    """
    list_domains = []
    for r in rules:
        list_domains.append(extract_domain(r))
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not rules:
        raise MissingArgumentError(
            f'{Fore.RED}ERROR: you need to specify at least one domain to delete from rules{Style.RESET_ALL}'
        )
    elif not force and not prompt_user('You are about to add rules for the following domains:', list_domains):
        raise SIGTERM()
    print(':: Deleting rules...')
    lock_database()
    for rule_data in rules:
        rule_obj = Rule(rule_data)
        try:
            rule_obj.delete()
        except RuleNotExistsError as err:
            raise RuleNotExistsError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    update_hosts(get_all_rules())
    unlock_database()


def list_rules(standard_only: bool = False, user_only: bool = False) -> None:
    """List all active rules and their policy

    Args:
        standard_only (bool, optional): Only list standard rules
        user_only (bool, optional): Only list user rules
    """
    if not standard_only and not user_only:
        rules = get_all_rules()
    elif standard_only and not user_only:
        rules = get_all_rules(user=False)
    elif user_only and not standard_only:
        rules = get_all_rules(standard=False)
    else:
        rules = get_all_rules(False, False)

    for r in rules.keys():
        if rules[r]['policy'] == 'redirect':
            print(f'REDIRECT    {r}')
        elif rules[r]['policy'] == 'block':
            print(f'BLOCK       {r}')
        elif rules[r]['policy'] == 'allow':
            print(f'ALLOW       {r}')



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import os.path
import re

# External libraries
from colorama import Fore, Style

# Local libraries
from ..exceptions import InvalidFilterSyntax, SIGTERM
from ..utils import prompt_user
from ..config import Font


class Syntax:
    """The class containing all supported filter formats
    """
    adblockplus = 'adblockplus'
    hosts = 'hosts'
    tblock = 'tblock'
    list = 'list'
    dnsmasq = 'dnsmasq'
    opera = 'opera'


def is_comment(line: str, syntax: str) -> bool:
    """Check if a line is a comment

    Args:
        line (str): The line to check
        syntax (str): The syntax to use for check

    Returns:
        bool: True if line is a comment
    """
    if hasattr(Syntax, syntax):
        if syntax == Syntax.adblockplus:
            return line[0:1] == '!'
        else:
            return line[0:1] == '#'
    else:
        raise InvalidFilterSyntax(f'"{syntax}" is not a valid syntax')


class FilterParser:

    def __init__(self, filter_path: str, syntax: str = None) -> None:
        """
        Filter to parse
        Args:
            filter_path (str): The path to the filter
            syntax (str, optional): The syntax of the filter (if not set, tblock will autodetect it)
        """
        self.filter_path = os.path.realpath(filter_path)

        # Prevent files that are not UTF-8 encoded to be opened as filters
        try:
            with open(self.filter_path, 'rt', encoding='utf-8') as f:
                self.list = f.readlines()
        except UnicodeDecodeError:
            raise InvalidFilterSyntax(f'cannot read file "{self.filter_path}" since its encoding is not supported')

        self.total_lines = len(self.list)

        if syntax:
            self.syntax = syntax
        else:
            try:
                self.syntax = self.detect_syntax()
            except ZeroDivisionError:
                self.syntax = ''

    def detect_syntax(self) -> str:
        """Detect the syntax of a filter

        Returns:
            str: The syntax. See the Syntax object for more information.
        """
        rules = []

        # Remove blank lines from test
        for rule in self.list:
            if rule != '\n':
                rules.append(rule.split("\n")[0])

        # Count the number of lines that are not empty
        rules_count = len(rules)

        # Convert that list into a string, in order to use regex to find patterns in it
        rules = str(rules)

        # Check the syntax of the filter using regex and calculating the percentage of rules that match a syntax
        if re.findall(re.compile(r'(\[adblock plus)', re.IGNORECASE), rules):
            return Syntax.adblockplus
        elif re.findall(re.compile(r'(@BEGIN_RULES|@END_RULES)'), rules):
            return Syntax.tblock
        elif len(re.findall(re.compile(r'(\|\|[.a-z\-]*[$^]|![ A-z]*[.]*:|!)'), rules)) * 100 / rules_count >= 50:
            return Syntax.adblockplus
        elif len(re.findall(re.compile(r'[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}[ \t]|::1[ \t]'), rules)) \
                * 100 / rules_count >= 50:
            return Syntax.hosts
        elif len(re.findall(re.compile(r'(server=/|domain=/)'), rules)) * 100 / rules_count >= 50:
            return Syntax.dnsmasq
        elif len(re.findall(re.compile(r'(http://)'), rules)) * 100 / rules_count >= 50:
            return Syntax.opera
        elif len(re.findall(re.compile(r'([0-9a-z]*\.[.a-z]*)'), rules)) * 100 / rules_count >= 50:
            return Syntax.list
        else:
            return ''

    def get_filter_content(self, allow: bool = True, block: bool = True,
                           redirect: bool = True, comments: bool = False) -> list:
        """Get all rules from a filter

        Args:
            allow (bool, optional): Return allowing rules when supported by syntax (default)
            block (bool, optional): Return blocking rules when supported by syntax(default)
            redirect (bool, optional): Return redirecting rules when supported by syntax (default)
            comments (bool, optional): Return comments when supported by syntax (False by default)
        """

        # Prevent unsupported syntax to be opened
        if not hasattr(Syntax, self.syntax):
            raise InvalidFilterSyntax(f'"{self.syntax}" is not a valid syntax')
        filter_data = []

        # Variables for specific filter formats
        tblock_begin = None
        tblock_policy = None
        opera_policy = None

        count = 0
        load_character = '|'

        print(f' [{Fore.BLUE}i{Style.RESET_ALL}] Filter syntax is: {Font.UNDERLINE}{self.syntax}{Font.DEFAULT}')

        for rule in self.list:
            if not count + 1 == self.total_lines:
                print(f' [{Fore.YELLOW}{load_character}{Style.RESET_ALL}] Parsing filter '
                      f'({count + 1}/{self.total_lines})', end='\r')
                # Show process
                if load_character == '|' and not count % 500:
                    load_character = '/'
                elif load_character == '/' and not count % 500:
                    load_character = '-'
                elif load_character == '-' and not count % 500:
                    load_character = '\\'
                elif load_character == '\\' and not count % 500:
                    load_character = '|'
            else:
                print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Parsing filter '
                      f'({count + 1}/{self.total_lines})')

            # Skip all blank lines
            if rule != "\n":
                if not is_comment(rule, self.syntax):

                    if self.syntax == Syntax.adblockplus:
                        if not re.findall(re.compile(r'([$/&%?:*()=,;#\"+_])'), rule):
                            if '^' in rule:
                                if rule[0:2] == '||' and block:
                                    filter_data.append(
                                        ['rule', rule.split("\n")[0].split('||')[1].split('^')[0], 'block']
                                    )
                                elif rule[0:4] == '@@||' and allow:
                                    filter_data.append(
                                        ['rule', rule.split("\n")[0].split('||')[1].split('^')[0], 'allow']
                                    )

                    elif self.syntax == Syntax.hosts:
                        if re.match(re.compile(r'(127\.0\.[01]\.[0-9]{1,3}[ \t]|0\.0\.0\.0[ \t]|::1[ \t])'), rule):
                            for k in rule.split('\n')[0].split('#')[0].split(' '):
                                for r in k.split('\t'):
                                    if not re.findall(re.compile(r'[$/&%?:*()=,;#\"+\t_]'), r):
                                        if not re.findall(
                                                re.compile(r'(127\.0\.[01]\.[0-9]{1,3}|0\.0\.0\.0)'), r
                                        ) and r and block:
                                            filter_data.append(
                                                ['rule', r, 'block']
                                            )
                        else:
                            if not re.findall(re.compile(r'([$/&%?:*()=,;#\"+_])'), rule):
                                ip = rule.split(' ')[0].split('\t')[0]
                                rule = rule.split('\n')[0].split(' ')[len(rule.split(' ')) - 1].replace(' ', '')
                                rule = rule.split('\t')[1] if '\t' in rule else rule
                                if rule and redirect:
                                    filter_data.append(
                                        ['rule', rule, 'redirect', ip]
                                    )

                    elif self.syntax == Syntax.list:
                        if not re.findall(re.compile(r'([$/&%?:*()=,;#\"+_])'), rule) and block:
                            filter_data.append(['rule', rule.split("\n")[0], 'block'])

                    elif self.syntax == Syntax.dnsmasq:
                        if re.match(re.compile(r'(server=/|domain=/)'), rule):
                            if not re.findall(re.compile(r'([$&%?:*(),;#\"+_])'), rule) and block:
                                filter_data.append(
                                    ['rule', rule.split("\n")[0].split('=/')[1].split('/')[0], 'block']
                                )

                    elif self.syntax == Syntax.tblock:
                        if tblock_begin == 'rules':
                            if rule.split('\n')[0] == '@END_RULES':
                                tblock_begin = None
                            elif rule[0:1] == '!':
                                if rule.split('\n')[0] == '!allow':
                                    tblock_policy = 'allow'
                                elif rule.split('\n')[0] == '!block':
                                    tblock_policy = 'block'
                                elif rule[0:10] == '!redirect ':
                                    tblock_policy = 'redirect ' + rule.split('!redirect ')[1].split('\n')[0]
                            else:
                                if not re.findall(re.compile(r'([$/&%?:*()=,;#\"+_])'), rule):
                                    if 'redirect' in tblock_policy and redirect:
                                        filter_data.append([
                                            'rule', rule.split("\n")[0], 'redirect',
                                            tblock_policy.split('redirect ')[1]
                                        ])
                                    elif tblock_policy == 'block' and block or tblock_policy == 'allow' and allow:
                                        filter_data.append(['rule', rule.split("\n")[0], tblock_policy])
                        elif tblock_begin is None:
                            if rule.split('\n')[0] == '@BEGIN_RULES':
                                tblock_begin = 'rules'

                    elif self.syntax == Syntax.opera:
                        if rule.split('\n')[0] == '[include]':
                            opera_policy = 'allow'
                        elif rule.split('\n')[0] == '[exclude]':
                            opera_policy = 'block'
                        if re.match(r'^http://[a-z0-9]*\.[a-z0-9.]*/\*', rule.split('\n')[0]):
                            if opera_policy == 'allow':
                                filter_data.append(['rule', rule.split("http://")[1].split('/*')[0], 'allow'])
                            elif opera_policy == 'block':
                                filter_data.append(['rule', rule.split("http://")[1].split('/*')[0], 'block'])

                elif comments:
                    if self.syntax == Syntax.adblockplus:
                        filter_data.append(['comment', rule.split('!')[1].replace('\n', '')])
                    else:
                        filter_data.append(['comment', rule.split('#')[1].replace('\n', '')])
            count += 1
        print(f' [{Fore.BLUE}i{Style.RESET_ALL}] Skipped {self.total_lines - len(filter_data)} '
              f'invalid rule(s) or blank line(s)')
        return filter_data

    def convert(self, output_file: str, output_syntax: str, hosts_default_ip: str = '127.0.0.1', allow: bool = True,
                block: bool = True, redirect: bool = True, comments: bool = False, force: bool = False,
                verbosity: bool = False) -> None:
        """Convert a filter into another syntax

        Args:
            output_file (str): Where to write converted filter
            output_syntax (str): To which syntax the filter should be converted
            hosts_default_ip (str, optional): Specify a custom IP where to redirect blocked domains in hosts file format
            allow (bool, optional): Convert allowing rules when supported by syntax (default)
            block (bool, optional): Convert blocking rules when supported by syntax (default)
            redirect (bool, optional): Convert redirecting rules when supported by syntax (default)
            comments (bool, optional): Convert comments when supported by syntax (False by default)
            force (bool, optional): Do not prompt for anything (False by default)
            verbosity (bool, optional): Enable verbosity
        """

        # If output file already exists that the force option is not set, and that the user answers
        # no to the confirm prompt, exit the script
        if os.path.isfile(os.path.realpath(output_file)) and \
                not force and not prompt_user(f'You are about to overwrite the file "{output_file}"'):
            raise SIGTERM()

        print(':: Converting filter...')

        if verbosity:
            print(f'{Font.BOLD}==> Converting into: {output_syntax}{Font.DEFAULT}')

        # Check if the syntax is supported, raise an exception if not
        if not hasattr(Syntax, output_syntax):
            raise InvalidFilterSyntax(f'"{output_syntax}" is not a valid syntax')

        # Set the first line for most filter formats
        if output_syntax == Syntax.adblockplus:
            output_content = '[Adblock Plus 3.1]\n\n! Converted by tblockc (https://tblock.codeberg.page/)\n\n'
        elif output_syntax == Syntax.tblock:
            output_content = '# Converted by tblockc (https://tblock.codeberg.page/)\n\n@BEGIN_RULES\n\n'
        elif not output_syntax == Syntax.list:
            output_content = '# Converted by tblockc (https://tblock.codeberg.page/)\n\n'
        else:
            output_content = ''

        # Variables for specific filter formats
        tblock_syntax = None
        opera_policy = None
        count = 1
        load_character = '|'

        for rule in self.get_filter_content(allow, block, redirect, comments):

            if count < self.total_lines:
                print(f' [{Fore.YELLOW}{load_character}{Style.RESET_ALL}] Converting rules '
                      f'({count}/{self.total_lines})', end='\r')
                # Show process
                if load_character == '|' and not count % 500:
                    load_character = '/'
                elif load_character == '/' and not count % 500:
                    load_character = '-'
                elif load_character == '-' and not count % 500:
                    load_character = '\\'
                elif load_character == '\\' and not count % 500:
                    load_character = '|'
            else:
                print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Converting rules '
                      f'({count}/{self.total_lines})')

            if rule[0] == 'comment' and comments:
                if output_syntax == Syntax.adblockplus:
                    output_content += f'!{rule[1]}\n'
                elif not output_syntax == Syntax.list:
                    output_content += f'#{rule[1]}\n'
            elif rule[0] == 'rule':

                if rule[2] == 'allow' and allow:
                    if output_syntax == Syntax.adblockplus:
                        output_content += f'@@||{rule[1]}^\n'
                    elif output_syntax == Syntax.tblock:
                        if tblock_syntax != 'allow':
                            tblock_syntax = 'allow'
                            output_content += f'\n!allow\n{rule[1]}\n'
                        else:
                            output_content += f'{rule[1]}\n'
                    elif output_syntax == Syntax.opera:
                        if opera_policy != 'allow':
                            opera_policy = 'allow'
                            output_content += f'[include]\nhttp://{rule[1]}/*\n'
                            output_content += f'http://*.{rule[1]}/*\n'
                        else:
                            output_content += f'http://{rule[1]}/*\n'
                            output_content += f'http://*.{rule[1]}/*\n'

                elif rule[2] == 'block' and block:
                    if output_syntax == Syntax.adblockplus:
                        output_content += f'||{rule[1]}^\n'
                    elif output_syntax == Syntax.dnsmasq:
                        output_content += f'server=/{rule[1]}/\n'
                    elif output_syntax == Syntax.hosts:
                        output_content += f'{hosts_default_ip}    {rule[1]}\n'
                    elif output_syntax == Syntax.list:
                        output_content += f'{rule[1]}\n'
                    elif output_syntax == Syntax.tblock:
                        if tblock_syntax != 'block':
                            tblock_syntax = 'block'
                            output_content += f'\n!block\n{rule[1]}\n'
                        else:
                            output_content += f'{rule[1]}\n'
                    elif output_syntax == Syntax.opera:
                        if opera_policy != 'block':
                            opera_policy = 'block'
                            output_content += f'[exclude]\nhttp://{rule[1]}/*\n'
                            output_content += f'http://*.{rule[1]}/*\n'
                        else:
                            output_content += f'http://{rule[1]}/*\n'
                            output_content += f'http://*.{rule[1]}/*\n'

                elif rule[2] == 'redirect' and redirect:
                    if output_syntax == Syntax.hosts:
                        output_content += f'{rule[3]}    {rule[1]}\n'
                    elif output_syntax == Syntax.tblock:
                        if tblock_syntax != f'redirect {rule[3]}':
                            tblock_syntax = f'redirect {rule[3]}'
                            output_content += f'\n!redirect {rule[3]}\n{rule[1]}\n'
                        else:
                            output_content += f'{rule[1]}\n'

            count += 1
        if output_syntax == Syntax.tblock:
            output_content += '\n@END_RULES\n'
        with open(os.path.realpath(output_file), 'wt') as f:
            f.write(output_content)
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Filter successfully converted')



#!/usr/bin/env python

import os
import tblock.config
import tblock.converter
import unittest

tblock.config.Path.HOSTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts')
tblock.config.Path.DATABASE = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'storage.sqlite')
tblock.config.Path.DB_LOCK = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', '.db-lock')
tblock.config.Path.TMP_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root')
tblock.config.Path.HOSTS_VERIFICATION = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'verification')
tblock.config.setup_database(tblock.config.Path.DATABASE)

test_custom_filter = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'filter.txt')

adblockplus = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'abp.txt')
hosts_127 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts.127')
hosts_0 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'hosts.0')
dnsmasq = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'dnsmasq.conf')
_list = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'list.txt')
tbf = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'tblock.tbf')
opera = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'root', 'opera.txt')


class TestConverter(unittest.TestCase):

    def test_convert_adblockplus_hosts(self):
        self.assertEquals(
            tblock.converter.convert_filter(test_custom_filter, hosts_127, 'hosts', force=True),
            None
        )

    def test_convert_adblockplus_hosts_0(self):
        self.assertEquals(
            tblock.converter.convert_filter(
                test_custom_filter, hosts_0, 'hosts', hosts_default_ip='0.0.0.0', force=True
            ),
            None
        )

    def test_convert_adblockplus_dnsmasq(self):
        self.assertEquals(
            tblock.converter.convert_filter(test_custom_filter, dnsmasq, 'dnsmasq', force=True),
            None
        )

    def test_convert_adblockplus_list(self):
        self.assertEquals(
            tblock.converter.convert_filter(test_custom_filter, _list, 'list', force=True),
            None
        )

    def test_convert_adblockplus_tblock(self):
        self.assertEquals(
            tblock.converter.convert_filter(test_custom_filter, tbf, 'tblock', force=True),
            None
        )

    def test_convert_adblockplus_opera(self):
        self.assertEquals(
            tblock.converter.convert_filter(test_custom_filter, opera, 'opera', force=True),
            None
        )

    def test_convert_hosts_adblockplus(self):
        self.assertEquals(
            tblock.converter.convert_filter(hosts_127, adblockplus, 'adblockplus', force=True),
            None
        )

    def test_convert_hosts_dnsmasq(self):
        self.assertEquals(
            tblock.converter.convert_filter(hosts_127, dnsmasq, 'dnsmasq', force=True),
            None
        )

    def test_convert_hosts_list(self):
        self.assertEquals(
            tblock.converter.convert_filter(hosts_127, _list, 'list', force=True),
            None
        )

    def test_convert_hosts_tblock(self):
        self.assertEquals(
            tblock.converter.convert_filter(hosts_127, tbf, 'tblock', force=True),
            None
        )

    def test_convert_hosts_opera(self):
        self.assertEquals(
            tblock.converter.convert_filter(hosts_127, opera, 'opera', force=True),
            None
        )

    def test_convert_dnsmasq_adblockplus(self):
        self.assertEquals(
            tblock.converter.convert_filter(dnsmasq, adblockplus, 'adblockplus', force=True),
            None
        )

    def test_convert_dnsmasq_hosts(self):
        self.assertEquals(
            tblock.converter.convert_filter(dnsmasq, hosts_127, 'hosts', force=True),
            None
        )

    def test_convert_dnsmasq_list(self):
        self.assertEquals(
            tblock.converter.convert_filter(dnsmasq, _list, 'list', force=True),
            None
        )

    def test_convert_dnsmasq_tblock(self):
        self.assertEquals(
            tblock.converter.convert_filter(dnsmasq, tbf, 'tblock', force=True),
            None
        )

    def test_convert_dnsmasq_opera(self):
        self.assertEquals(
            tblock.converter.convert_filter(dnsmasq, opera, 'opera', force=True),
            None
        )

    def test_convert_list_adblockplus(self):
        self.assertEquals(
            tblock.converter.convert_filter(_list, adblockplus, 'adblockplus', force=True),
            None
        )

    def test_convert_list_hosts(self):
        self.assertEquals(
            tblock.converter.convert_filter(_list, hosts_127, 'hosts', force=True),
            None
        )

    def test_convert_list_dnsmasq(self):
        self.assertEquals(
            tblock.converter.convert_filter(_list, dnsmasq, 'dnsmasq', force=True),
            None
        )

    def test_convert_list_tblock(self):
        self.assertEquals(
            tblock.converter.convert_filter(_list, tbf, 'tblock', force=True),
            None
        )

    def test_convert_list_opera(self):
        self.assertEquals(
            tblock.converter.convert_filter(_list, opera, 'opera', force=True),
            None
        )

    def test_convert_tblock_adblockplus(self):
        self.assertEquals(
            tblock.converter.convert_filter(tbf, adblockplus, 'adblockplus', force=True),
            None
        )

    def test_convert_tblock_hosts(self):
        self.assertEquals(
            tblock.converter.convert_filter(tbf, hosts_127, 'hosts', force=True),
            None
        )

    def test_convert_tblock_dnsmasq(self):
        self.assertEquals(
            tblock.converter.convert_filter(tbf, dnsmasq, 'dnsmasq', force=True),
            None
        )

    def test_convert_tblock_list(self):
        self.assertEquals(
            tblock.converter.convert_filter(tbf, _list, 'list', force=True),
            None
        )

    def test_convert_tblock_opera(self):
        self.assertEquals(
            tblock.converter.convert_filter(tbf, opera, 'opera', force=True),
            None
        )

    def test_convert_opera_adblockplus(self):
        self.assertEquals(
            tblock.converter.convert_filter(opera, adblockplus, 'adblockplus', force=True),
            None
        )

    def test_convert_opera_hosts(self):
        self.assertEquals(
            tblock.converter.convert_filter(opera, hosts_127, 'hosts', force=True),
            None
        )

    def test_convert_opera_dnsmasq(self):
        self.assertEquals(
            tblock.converter.convert_filter(opera, dnsmasq, 'dnsmasq', force=True),
            None
        )

    def test_convert_opera_list(self):
        self.assertEquals(
            tblock.converter.convert_filter(opera, _list, 'list', force=True),
            None
        )

    def test_convert_opera_tblock(self):
        self.assertEquals(
            tblock.converter.convert_filter(opera, tbf, 'tblock', force=True),
            None
        )

    def test_syntax_detection(self):
        self.assertEquals(
            tblock.converter.get_syntax(test_custom_filter),
            None
        )


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import os.path
import sys
import re
import sqlite3

# External libraries
from colorama import Fore, Style

# Local libraries
from .tools import enable_protection, disable_protection, print_status, print_version_info, print_help
from .config import check_dirs
from .repo import sync_remote_repo
from .rules import add_rules, list_rules, delete_rules
from .converter import convert_filter, get_syntax
from .utils import check_root, unlock_database
from .filters import subscribe_to_filters, show_filters, update_filters, update_all_filters, unsubscribe_from_filters, \
    list_filters, change_filters_permissions, change_filter_id, search_filters
from .exceptions import FilterNotExists, MissingArgumentError, FilterSourceExistsError, NotSubscribingError, \
    RuleNotExistsError, InvalidBindAddress, SIGTERM, InvalidFilterSyntax, NetworkError, FilterNotCustomError, \
    FilterExistsError


tblock_args = [
    '-h', '--help',
    '-s', '--status',
    '-v', '--version',
    '-a', '--allow',
    '-b', '--block',
    '-r', '--redirect',
    '-d', '--delete-rule',
    '-l', '--list-rules',
    '-S', '--subscribe',
    '-C', '--add-custom',
    '-R', '--remove',
    '-U', '--update',
    '-Q', '--search',
    '-M', '--mod',
    '-X', '--change-id',
    '-I', '--info',
    '-L', '--list',
    '-Y', '--sync',
    '-H', '--update-hosts',
    '-D', '--restore'
]

tblock_opts = [
    '-f', '--force',
    '-t', '--standard',
    '-u', '--user',
    '-y', '--sync',
    '-c', '--custom',
    '-k', '--subscribing',
    '-n', '--unsubscribing',
    '-w', '--on-remote-repo',
]


tblockc_opts = [
    '-h', '--help',
    '-f', '--force',
    '-v', '--version',
    '-c', '--comments',
    '-o', '--output',
    '-i', '--input-syntax',
    '-s', '--syntax',
    '-g', '--get-syntax',
    '-0', '--zero'
]


class TBlockArguments:

    def __init__(self, args: list) -> None:
        """Arguments object

        Args:
            args (list): A list of all arguments
        """
        self.action = None
        self.options = {
            'force': False,
            'standard': False,
            'user': False,
            'custom': False,
            'sync': False,
            'subscribing': False,
            'unsubscribing': False,
            'on-remote-repo': False
        }
        self.values = []

        for argument in args:

            if re.match(re.compile(r'--'), argument):
                # Is an argument
                if argument in tblock_args:
                    self.action = argument.split('--')[1]

                # Is an option
                elif argument in tblock_opts:
                    self.options[argument.split('--')[1]] = True

                # Other
                else:
                    print(f'{Fore.RED}ERROR: invalid argument "{argument}"{Style.RESET_ALL}')
                    sys.exit(1)

            elif re.match(re.compile(r'-'), argument):
                # Is an argument
                if argument in tblock_args:
                    self.action = tblock_args[tblock_args.index(argument) + 1].split('--')[1]

                # Is an option
                elif argument in tblock_opts:
                    self.options[tblock_opts[tblock_opts.index(argument) + 1].split('--')[1]] = True

                # Contains an argument
                elif argument[0:2] in tblock_args:
                    self.action = tblock_args[tblock_args.index(argument[0:2]) + 1].split('--')[1]
                    for option in argument[2:]:
                        if f'-{option}' in tblock_opts:
                            self.options[tblock_opts[tblock_opts.index(f'-{option}') + 1].split('--')[1]] = True
                        else:
                            print(f'{Fore.RED}ERROR: invalid option "{option}"{Style.RESET_ALL}')
                            sys.exit(1)

                # Contains an option
                elif argument[0:2] in tblock_opts:
                    self.options[tblock_opts[tblock_opts.index(argument[0:2]) + 1].split('--')[1]] = True
                    for option in argument[2:]:
                        if f'-{option}' in tblock_opts:
                            self.options[tblock_opts[tblock_opts.index(f'-{option}') + 1].split('--')[1]] = True
                        else:
                            print(f'{Fore.RED}ERROR: invalid option "{option}"{Style.RESET_ALL}')
                            sys.exit(1)

                # Other
                else:
                    print(f'{Fore.RED}ERROR: invalid argument "{argument}"{Style.RESET_ALL}')
                    sys.exit(1)

            else:
                self.values.append(argument)

        if not self.action:
            print(f'{Fore.RED}ERROR: one argument is required{Style.RESET_ALL}\nTo show the help page, run "tblock -h"')
            sys.exit(1)


class TBlockcArguments:

    def __init__(self, args: list) -> None:
        """Arguments object

        Args:
            args (list): A list of all arguments
        """
        self.options = {
            'help': [bool, False],
            'force': [bool, False],
            'version': [bool, False],
            'comments': [bool, False],
            'output': [str, None],
            'syntax': [str, None],
            'input-syntax': [str, None],
            'get-syntax': [bool, False],
            'zero': [bool, False]
        }

        self.filter_path = None

        for argument in args:

            if re.match(re.compile(r'--'), argument):
                # Is an option
                if argument in tblockc_opts:
                    if self.options[argument.split('--')[1]][0] is bool:
                        self.options[argument.split('--')[1]][1] = True
                    elif self.options[argument.split('--')[1]][0] is str:
                        try:
                            self.options[argument.split('--')[1]][1] = args[args.index(argument) + 1]
                            args.pop(args.index(argument) + 1)
                        except IndexError:
                            print(f'{Fore.RED}ERROR: option "{argument}" requires a value{Style.RESET_ALL}')
                            sys.exit(1)

                # Other
                else:
                    print(f'{Fore.RED}ERROR: invalid option "{argument}"{Style.RESET_ALL}')
                    sys.exit(1)

            elif re.match(re.compile(r'-'), argument):
                # Is an option
                if argument in tblockc_opts:
                    if self.options[tblockc_opts[tblockc_opts.index(argument) + 1].split('--')[1]][0] is bool:
                        self.options[tblockc_opts[tblockc_opts.index(argument) + 1].split('--')[1]][1] = True
                    elif self.options[tblockc_opts[tblockc_opts.index(argument) + 1].split('--')[1]][0] is str:
                        try:
                            self.options[tblockc_opts[tblockc_opts.index(argument) + 1].split('--')[1]][1] = \
                                args[args.index(argument) + 1]
                            args.pop(args.index(argument) + 1)
                        except IndexError:
                            print(f'{Fore.RED}ERROR: option "{argument[0:2]}" requires a value{Style.RESET_ALL}')
                            sys.exit(1)

                # Contains an option
                elif argument[0:2] in tblockc_opts:

                    if self.options[tblockc_opts[tblockc_opts.index(argument[0:2]) + 1].split('--')[1]][0] is bool:
                        self.options[tblockc_opts[tblockc_opts.index(argument[0:2]) + 1].split('--')[1]][1] = True
                    else:
                        print(f'{Fore.RED}ERROR: option "{argument[0:2]}" requires a value{Style.RESET_ALL}')
                        sys.exit(1)

                    for option in argument[2:]:
                        if f'-{option}' in tblockc_opts:

                            if self.options[tblockc_opts[tblockc_opts.index(f'-{option}') + 1].split('--')[1]][0] \
                                    is bool:
                                self.options[tblockc_opts[tblockc_opts.index(f'-{option}') + 1].split('--')[1]][1] \
                                    = True

                            elif self.options[tblockc_opts[tblockc_opts.index(f'-{option}') + 1].split('--')[1]][0] \
                                    is str and len(argument[2:]) - 1 == argument[2:].index(option):
                                try:
                                    self.options[tblockc_opts[tblockc_opts.index(f'-{option}') + 1].split('--')[1]][1] \
                                        = args[args.index(argument) + 1]
                                    args.pop(args.index(argument) + 1)
                                except IndexError:
                                    print(
                                        f'{Fore.RED}ERROR: option "{option}" requires a value{Style.RESET_ALL}'
                                    )
                                    sys.exit(1)
                            else:
                                print(f'{Fore.RED}ERROR: option "{option}" requires a value{Style.RESET_ALL}')
                                sys.exit(1)

                        else:
                            print(f'{Fore.RED}ERROR: invalid option "{option}"{Style.RESET_ALL}')
                            sys.exit(1)
                # Other
                else:
                    print(f'{Fore.RED}ERROR: invalid option "{argument}"{Style.RESET_ALL}')
                    sys.exit(1)
            else:
                self.filter_path = argument


def parse_args_tblock() -> None:
    """Parse arguments and run the ad-blocker
    """
    parser = TBlockArguments(sys.argv[1:])

    if parser.action == 'help':
        print_help(True)

    elif parser.action == 'status':
        print_status()

    elif parser.action == 'version':
        print_version_info(True)

    elif parser.action == 'allow':
        data = []
        for rule in parser.values:
            data.append([rule, 'allow'])
        add_rules(data, parser.options['force'])

    elif parser.action == 'block':
        data = []
        for rule in parser.values:
            data.append([rule, 'block'])
        add_rules(data, parser.options['force'])

    elif parser.action == 'redirect':
        data = []
        for rule in parser.values[:len(parser.values) - 1]:
            data.append([rule, 'redirect', parser.values[len(parser.values) - 1]])
        add_rules(data, parser.options['force'])

    elif parser.action == 'delete-rule':
        delete_rules(parser.values, parser.options['force'])

    elif parser.action == 'list-rules':
        list_rules(parser.options['standard'], parser.options['user'])

    elif parser.action == 'subscribe':
        data = []
        for filter_id in parser.values:
            data.append([filter_id, 'b'])
        if parser.options['sync']:
            sync_remote_repo(parser.options['force'])
            print()
        subscribe_to_filters(data, parser.options['force'])

    elif parser.action == 'add-custom':
        data = []
        for filter_id in parser.values:
            data.append([filter_id, 'b', os.path.realpath(parser.values[parser.values.index(filter_id) + 1])])
            parser.values.pop(parser.values.index(filter_id) + 1)
        if parser.options['sync']:
            sync_remote_repo(parser.options['force'])
            print()
        subscribe_to_filters(data, parser.options['force'])

    elif parser.action == 'remove':
        if parser.options['sync']:
            sync_remote_repo(parser.options['force'])
            print()
        unsubscribe_from_filters(parser.values, parser.options['force'])

    elif parser.action == 'update':
        if parser.options['sync']:
            sync_remote_repo(parser.options['force'])
            print()
        if parser.values:
            update_filters(parser.values, parser.options['force'])
        else:
            update_all_filters(parser.options['force'])

    elif parser.action == 'search':
        for query in parser.values:
            search_filters(query, parser.options['on-remote-repo'], parser.options['custom'],
                           parser.options['subscribing'], parser.options['unsubscribing'])

    elif parser.action == 'mod':
        data = []
        try:
            permissions = parser.values[0]
            for filter_id in parser.values[1:]:
                data.append([filter_id, permissions])
            if parser.options['sync']:
                sync_remote_repo(parser.options['force'])
                print()
            change_filters_permissions(data, parser.options['force'])
        except IndexError:
            if not check_root():
                raise PermissionError(
                    f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
            else:
                raise MissingArgumentError(
                    f'{Fore.RED}ERROR: please specify the new policy and the filter to change{Style.RESET_ALL}'
                )

    elif parser.action == 'change-id':
        # This verification should stay here, as only two values are required and not a list
        if not check_root():
            raise PermissionError(
                f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
        if len(parser.values) > 2:
            raise IOError(f'{Fore.RED}ERROR: too much arguments are given for changing filter ID{Style.RESET_ALL}')
        if parser.options['sync']:
            sync_remote_repo(parser.options['force'])
            print()
        try:
            change_filter_id(parser.values[0], parser.values[1])
        except IndexError:
            raise MissingArgumentError(
                f'{Fore.RED}ERROR: you need to specify the filter to rename as well as its new ID{Style.RESET_ALL}'
            )

    elif parser.action == 'info':
        show_filters(parser.values)

    elif parser.action == 'list':
        list_filters(parser.options['on-remote-repo'], parser.options['custom'],
                     parser.options['subscribing'], parser.options['unsubscribing'])

    elif parser.action == 'sync':
        sync_remote_repo(parser.options['force'])

    elif parser.action == 'update-hosts':
        enable_protection(parser.options['force'])

    elif parser.action == 'restore':
        disable_protection(parser.options['force'])

    else:
        pass    # This should never happen


def parse_args_tblockc() -> None:
    """Parse arguments and run the converter
    """
    parser = TBlockcArguments(sys.argv[1:])

    if parser.options['help'][1]:
        print_help(False)

    elif parser.options['version'][1]:
        print_version_info(False)

    else:

        if not parser.filter_path:
            print(f'{Fore.RED}ERROR: please specify at least one file{Style.RESET_ALL}\n'
                  f'To show the help page, run "tblockc -h"')
            sys.exit(1)

        elif parser.options['get-syntax'][1]:
            get_syntax(parser.filter_path)

        else:
            if not parser.options['output'][1]:
                print(f'{Fore.RED}ERROR: you must specify the output filter{Style.RESET_ALL}')
                sys.exit(1)

            elif not parser.options['syntax'][1]:
                print(f'{Fore.RED}ERROR: you must specify the output syntax{Style.RESET_ALL}')
                sys.exit(1)

            else:
                default_ip = '0.0.0.0' if parser.options['zero'][1] else '127.0.0.1'

                convert_filter(
                    input_file=parser.filter_path,
                    output_file=parser.options['output'][1],
                    output_syntax=parser.options['syntax'][1],
                    hosts_default_ip=default_ip,
                    input_syntax=parser.options['input-syntax'][1],
                    comments=parser.options['comments'][1],
                    force=parser.options['force'][1]
                )


def run() -> None:
    check_dirs()
    try:
        parse_args_tblock()
    except (PermissionError, FilterSourceExistsError, MissingArgumentError, FilterNotExists, NotSubscribingError,
            RuleNotExistsError, InvalidBindAddress, InvalidFilterSyntax, FileNotFoundError, NetworkError,
            FilterExistsError, FilterNotCustomError) as err:
        print(str(err).replace('[Errno 2] N', 'n'))
        unlock_database()
        sys.exit(1)
    except sqlite3.OperationalError:
        print(f'{Fore.RED}ERROR: database error: please run "tblock -Yf" as superuser '
              f'to resolve this problem{Style.RESET_ALL}')
        sys.exit(1)
    except sqlite3.DatabaseError:
        print(f'{Fore.RED}ERROR: database may be locked or corrupted: please run "tblock -Yf" as superuser')
    except (SIGTERM, KeyboardInterrupt):
        unlock_database()
        sys.exit(1)
    else:
        unlock_database()


def run_converter() -> None:
    try:
        parse_args_tblockc()
    except (PermissionError, FilterSourceExistsError, MissingArgumentError, FilterNotExists, NotSubscribingError,
            RuleNotExistsError, InvalidBindAddress, InvalidFilterSyntax, FileNotFoundError) as err:
        print(str(err).replace('[Errno 2] N', 'n'))
        sys.exit(1)
    except (SIGTERM, KeyboardInterrupt):
        exit(1)



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import os.path

# External libraries
from colorama import Fore, Style

# Local libraries
from ..config import Font
from ..exceptions import InvalidFilterSyntax
from .parser import FilterParser


def get_syntax(file: str) -> None:
    """Get the syntax of a filter

    Args:
        file (str): The path to the filter
    """
    print(':: Detecting syntax...')
    try:
        f_parser = FilterParser(os.path.realpath(file))
    except InvalidFilterSyntax as err:
        raise InvalidFilterSyntax(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    except FileNotFoundError as err:
        raise FileNotFoundError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    else:
        print(f' [{Fore.BLUE}i{Style.RESET_ALL}] Filter syntax is: {Font.UNDERLINE}{f_parser.syntax}{Font.DEFAULT}')


def convert_filter(input_file: str, output_file: str, output_syntax: str,
                   input_syntax: str = None, hosts_default_ip: str = '127.0.0.1', comments: bool = False,
                   allow: bool = True, block: bool = True, redirect: bool = True, force: bool = False) -> None:
    """Convert a filter into another syntax

    Args:
        input_file (str): The path to the filter to convert
        output_file (str): Where to write converted filter
        output_syntax (str): To which syntax the filter should be converted
        input_syntax (str, optional): The syntax of the filter (if not set, tblock will autodetect it)
        hosts_default_ip (str, optional): Specify a custom IP where to redirect blocked domains in hosts file format
        allow (bool, optional): Convert allowing rules when supported by syntax (default)
        block (bool, optional): Convert blocking rules when supported by syntax (default)
        redirect (bool, optional): Convert redirecting rules when supported by syntax (default)
        comments (bool, optional): Convert comments when supported by syntax (False by default)
        force (bool, optional): Do not prompt for anything (False by default)
    """
    try:
        f_parser = FilterParser(input_file, input_syntax)
    except InvalidFilterSyntax as err:
        raise InvalidFilterSyntax(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    except FileNotFoundError as err:
        raise FileNotFoundError(f'{Fore.RED}ERROR: {err}{Style.RESET_ALL}')
    else:
        f_parser.convert(output_file, output_syntax, hosts_default_ip, allow, block, redirect, comments, force, True)



# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__license__ = 'GPLv3'
__version__ = '1.1.1'

run_help = '''TBlock - An anticapitalist ad-blocker that uses the hosts file
Copyright (c) 2021 Twann <twann@ctemplar.com>

usage: tblock <operation> <options>

<operations>
* General
  -h  --help                              Show this help page
  -s  --status                            Show status information
  -v  --version                           Show version information
* Rules
  -a  --allow        [domain]             Allow a domain
  -b  --block        [domain]             Block a domain
  -r  --redirect     [domain]  [ip]       Redirect a domain to another address
  -d  --delete-rule  [domain]             Delete a rule for a domain
  -l  --list-rules   [options]            List active rules
* Filters
  -S  --subscribe    [filter]             Subscribe to a new filter
  -C  --add-custom   [id]     [file/url]  Subscribe to a custom filter
  -R  --remove       [filter]             Remove a filter (unsubscribe)
  -U  --update       [filter]             Update a filter
  -Q  --search       [query]  [options]   Search a query in the filter lists
  -M  --mod          [a|b|r]  [filter]    Change permissions of a filter
  -X  --change-id    [filter] [new-id]    Change the ID of a custom filter
  -I  --info         [filter]             Show information about a filter
  -L  --list         [options]            List all filters
  -Y  --sync                              Update remote filters repository
* Hosts
  -H  --update-hosts                      Update hosts file
  -D  --restore                           Restore default hosts file

<options>
* General options
  -f  --force                             Do not prompt user for anything
* Filters options
  -y  --sync                              Update remote filters repository
* List options
  -t  --standard                          Only show standard rules
  -u  --user                              Only show user rules
  -c  --custom                            Only show custom filters
  -k  --subscribing                       Only show subscribed filters
  -n  --unsubscribing                     Only show unsubscribed filters
  -w  --on-remote-repo                    Only show filters available on remote repository

For more information, see tblock(1)'''

run_converter_help = '''TBlockc - TBlock's built-in filter converter
Copyright (c) 2021 Twann <twann@ctemplar.com>

usage: tblockc [file] {-s,--syntax} [syntax] {-o,--output} [output_file] <options>

<options>
* General
  -h  --help                              Show this help page
  -f  --force                             Do not prompt user for anything
  -v  --version                           Show version information
* Converting
  -0  --zero                              Redirect domains to 0.0.0.0 instead of 127.0.0.1 (hosts)
  -c  --comments                          Also convert comments
  -o  --output       [file]               Original syntax of the filter to convert (required)
  -s  --syntax       [syntax]             New syntax to use (required)
  -i  --input-syntax [syntax]             Specify the syntax of the filter
* Other
  -g  --get-syntax   [file]               Scan a filter to get its syntax

<syntax>
  adblockplus                             AdBlock Plus syntax
  hosts                                   Hosts file format
  dnsmasq                                 dnsmasq.conf syntax
  list                                    Simple domain blacklist
  tblock                                  TBlock filter syntax
  opera                                   Opera filter syntax

For more information, see tblockc(1)'''



# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import sys
import sqlite3
import os.path

# External libraries
from colorama import Fore, Style

# Local libraries
from . import run_help, run_converter_help, __version__
from .config import Path, PROTECTING_STATUS, Font
from .rules import get_all_rules
from .utils import check_root, prompt_user
from .hosts import update_hosts, restore_hosts
from .exceptions import SIGTERM


def enable_protection(force: bool = False) -> None:
    """Update the hosts file and enable protection if it is not active

    Args:
        force (bool, optional): Do not prompt user (False by default)
    """
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to update the hosts file'):
        raise SIGTERM()
    else:
        print(':: Updating hosts file...')
        update_hosts(get_all_rules())


def disable_protection(force: bool = False) -> None:
    """Restore the default hosts file and disable the protection

    Args:
        force (bool, optional): Do not prompt user (False by default)
    """
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')
    elif not force and not prompt_user('You are about to restore the default hosts file and disable the protection'):
        raise SIGTERM()
    else:
        print(':: Disabling protection...')
        restore_hosts()


def print_status() -> None:
    """Print the current status of the ad-blocker
    """

    if PROTECTING_STATUS:
        status = f'{Fore.GREEN}blocking{Style.RESET_ALL}'
    else:
        status = f'{Fore.RED}not blocking{Style.RESET_ALL}'

    try:
        with sqlite3.connect(Path.DATABASE) as db:
            total_rules = db.cursor().execute('SELECT COUNT() FROM rules;').fetchone()[0]
            allow = db.cursor().execute('SELECT COUNT() FROM rules WHERE policy="allow";').fetchone()[0]
            block = db.cursor().execute('SELECT COUNT() FROM rules WHERE policy="block";').fetchone()[0]
            redirect = db.cursor().execute('SELECT COUNT() FROM rules WHERE policy="redirect";').fetchone()[0]
            user = db.cursor().execute('SELECT COUNT() FROM rules WHERE priority=1;').fetchone()[0]
            total_filters = db.cursor().execute('SELECT COUNT() FROM filters WHERE subscribing=1;').fetchone()[0]
            custom = db.cursor().execute('SELECT COUNT() FROM filters WHERE on_rfr=0;').fetchone()[0]
            allow_redirect = db.cursor().execute(
                'SELECT COUNT() FROM filters WHERE permissions LIKE "%r%";').fetchone()[0]
            if os.path.isfile(Path.DB_LOCK):
                db_status = f'{Fore.YELLOW}locked{Style.RESET_ALL}'
            else:
                db_status = f'{Fore.GREEN}healthy{Style.RESET_ALL}'
    except (sqlite3.DatabaseError, sqlite3.OperationalError):
        total_rules = '-'
        allow = '-'
        block = '-'
        redirect = '-'
        user = '-'
        total_filters = '-'
        custom = '-'
        allow_redirect = '-'
        db_status = f'{Fore.RED}corrupted{Style.RESET_ALL}'

    if not allow_redirect:
        redirect_perm = f'{Fore.GREEN}0 filter(s) allowed{Style.RESET_ALL}'
    elif str(allow_redirect) != '-' and allow_redirect < 4:
        redirect_perm = f'{Fore.YELLOW}{allow_redirect} filter(s) allowed{Style.RESET_ALL}'
    else:
        redirect_perm = f'{Fore.RED}{allow_redirect} filter(s) allowed{Style.RESET_ALL}'

    if not PROTECTING_STATUS:
        hijack = f'{Fore.YELLOW}tblock is not active{Style.RESET_ALL}'
    else:
        try:
            with open(Path.HOSTS_VERIFICATION, 'rt') as h:
                verification_hosts = h.read()
        except FileNotFoundError:
            verification_hosts = ''
        with open(Path.HOSTS, 'rt') as h:
            hosts_real = h.read()
        if verification_hosts == hosts_real:
            hijack = f'{Fore.GREEN}no threat detected{Style.RESET_ALL}'
        else:
            hijack = f'{Fore.RED}threat(s) detected, run "tblock -H" to fix{Style.RESET_ALL}'

    try:
        with open(Path.REPO_VERSION, 'rt') as f:
            repo_version = f.read()
    except FileNotFoundError:
        repo_version = 0

    if os.path.isfile(Path.NEEDS_UPDATE):
        needs_update = f'{Fore.RED}needed{Style.RESET_ALL}'
    else:
        needs_update = f'{Fore.GREEN}not needed{Style.RESET_ALL}'

    if Path.PREFIX == '/data/data/com.termux/files/usr/lib/tblock/':
        _platform = 'android'
    else:
        _platform = sys.platform

    print(f'{Font.UNDERLINE}Status{Font.DEFAULT}')
    print(f'  TBlock       : version {__version__}')
    print(f'  Platform     : {_platform}')
    print(f'  Protection   : {status}')
    print(f'  Repository   : version {repo_version}')
    print(f'  Database     : {db_status}\n')
    print(f'{Font.UNDERLINE}Rules{Font.DEFAULT}')
    print(f'  Total        : {total_rules}')
    print(f'  User         : {user}')
    print(f'  Allowing     : {allow}')
    print(f'  Blocking     : {block}')
    print(f'  Redirecting  : {redirect}\n')
    print(f'{Font.UNDERLINE}Filters{Font.DEFAULT}')
    print(f'  Active       : {total_filters}')
    print(f'  Custom       : {custom}')
    print(f'  Full update  : {needs_update}\n')
    print(f'{Font.UNDERLINE}Security{Font.DEFAULT}')
    print(f'  Redirecting  : {redirect_perm}')
    print(f'  Hosts hijack : {hijack}')


def print_version_info(ad_blocker: bool) -> None:
    """Print the version of the program

    Args:
        ad_blocker (bool): If True, show info for tblock. If False, show info for tblockc
    """

    if ad_blocker:
        name = 'TBlock'
        desc = 'An anticapitalist ad-blocker that uses the hosts file'
    else:
        name = 'TBlockc'
        desc = 'TBlock\'s built-in filter converter'

    print(
        f'{name} version {__version__} - {desc}\n'
        'Copyright (C) 2021 Twann <twann@ctemplar.com>\n\n'
        'This program is free software: you can redistribute it and/or modify\n'
        'it under the terms of the GNU General Public License as published by\n'
        'the Free Software Foundation, either version 3 of the License, or\n'
        '(at your option) any later version.\n\n'
        'This program is distributed in the hope that it will be useful,\n'
        'but WITHOUT ANY WARRANTY; without even the implied warranty of\n'
        'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n'
        'GNU General Public License for more details.\n'
        'You should have received a copy of the GNU General Public License\n'
        'along with this program.  If not, see <https://www.gnu.org/licenses/>.'
    )


def print_help(ad_blocker: bool) -> None:
    """Print the help page of the program

    Args:
        ad_blocker (bool): If True, show info for tblock. If False, show info for tblockc
    """
    if ad_blocker:
        print(run_help)
    else:
        print(run_converter_help)



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import json
import os.path
import sqlite3
from xml.etree import ElementTree

# External libraries
import requests
import urllib3
from colorama import Fore, Style

# Local libraries
from .filters import Filter, get_all_from_rfr
from .config import Path, Var, Font
from .utils import download_file, check_root, lock_database, unlock_database
from .exceptions import InvalidLinkError, NetworkError, SIGTERM


def sync_remote_repo(force: bool = False) -> None:
    """Sync the remote filter repository

    Args:
        force (bool, optional): Force update the remote repository (disabled by default)
    """
    if not check_root():
        raise PermissionError(f'{Fore.RED}ERROR: you need to run as root to perform this operation{Style.RESET_ALL}')

    print(':: Updating remote repository...')
    lock_database()

    # Download index
    tmp_index = os.path.join(Path.TMP_DIR, 'remote-repo.xml.tmp')
    for mirror in Var.REPO_MIRRORS:
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Get mirror: {mirror}', end='\r')
        try:
            download_file(mirror, tmp_index)
        except (requests.exceptions.ConnectionError, urllib3.exceptions.MaxRetryError, ConnectionRefusedError,
                urllib3.exceptions.NewConnectionError, InvalidLinkError):
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Get mirror: {mirror}')
            if Var.REPO_MIRRORS.index(mirror) == len(Var.REPO_MIRRORS) - 1:
                raise NetworkError(f'{Fore.RED}ERROR: cannot download any remote repository mirror{Style.RESET_ALL}')
        else:
            print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Get mirror: {mirror}')
            break

    # Check if the repository is an XML file, and if it is a valid repository
    print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Checking repository validity', end='\r')
    try:
        data = ElementTree.parse(tmp_index).getroot()
    except ElementTree.ParseError:
        print(f' [{Fore.RED}x{Style.RESET_ALL}] Checking repository validity')
        print(f' [{Fore.RED}x{Style.RESET_ALL}] Failed to update remote repository')
        raise SIGTERM()
    else:
        if data.tag == 'repository':
            print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Checking repository validity')
        else:
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Checking repository validity')
            print(f' [{Fore.RED}x{Style.RESET_ALL}] Failed to update remote repository')
            raise SIGTERM()

    # Check repository version
    if os.path.isfile(Path.REPO_VERSION):
        with open(Path.REPO_VERSION, 'rt') as f:
            current_version = int(f.read())
    else:
        current_version = 0
    if len(str(current_version)) >= 6 and int(str(current_version)[0:6]) >= int(data.attrib['version']) and not force:
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Remote repository is up-to-date')
    else:

        # Upgrade the repository
        print(f' [{Fore.BLUE}i{Style.RESET_ALL}] Upgrading repository version {data.attrib["version"]} '
              f'over {str(current_version)}')
        all_filters = []
        count = 1
        db = sqlite3.connect(Path.DATABASE)
        for _filter in data:
            all_filters.append(_filter.attrib['id'])
        len_filters = len(all_filters)
        load_character = '|'
        warns = []

        for _filter in data:

            if count < len_filters:
                print(f' [{Fore.YELLOW}{load_character}{Style.RESET_ALL}] Updating filter index '
                      f'({count}/{len_filters})', end='\r')
                # Show process
                if load_character == '|' and not count % 2:
                    load_character = '/'
                elif load_character == '/' and not count % 2:
                    load_character = '-'
                elif load_character == '-' and not count % 2:
                    load_character = '\\'
                elif load_character == '\\' and not count % 2:
                    load_character = '|'
            else:
                print(
                    f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Updating filter index ({count}/{len_filters})'
                )

            metadata = {
                "title": None,
                "description": None,
                "homepage": None,
                "license": None,
                "syntax": None
            }
            new_filter = Filter(_filter.attrib['id'])

            if new_filter.exists and new_filter.on_rfr:
                source = new_filter.source
            else:
                source = None

            for attr in _filter:

                if attr.tag == 'title':
                    metadata['title'] = attr.text
                elif attr.tag == 'desc':
                    metadata['description'] = attr.text
                elif attr.tag == 'homepage':
                    metadata['homepage'] = attr.text
                elif attr.tag == 'license':
                    metadata['license'] = attr.text
                elif attr.tag == 'syntax':
                    metadata['syntax'] = attr.text
                elif attr.tag == 'source':
                    source = attr.text
            if new_filter.exists and new_filter.on_rfr:
                db.cursor().execute(
                    'UPDATE "filters" SET source=?, metadata=? WHERE id=?;',
                    (source, json.dumps(metadata), new_filter.id)
                )
            elif not new_filter.exists:
                db.cursor().execute(
                    'INSERT INTO "filters" ("id", "source", "metadata", "on_rfr", "subscribing") '
                    'VALUES (?, ?, ?, ?, ?)',
                    (new_filter.id, source, json.dumps(metadata), True, False)
                )
            elif new_filter.exists and not new_filter.on_rfr:
                warns.append(new_filter.id)
            else:
                pass

            count += 1
        db.commit()

        # Check for all filters that were removed from the remote repository
        all_filters_in_db = get_all_from_rfr()
        for filter_id in all_filters_in_db:
            if filter_id not in all_filters:
                new_filter = Filter(filter_id)
                if new_filter.subscribing:
                    # Transform the filter into a custom filter if the user is subscribing to it
                    db.cursor().execute(
                        'UPDATE "filters" SET source=?, metadata=?, on_rfr=? WHERE id=?;',
                        (new_filter.source, json.dumps(new_filter.metadata), False, new_filter.id)
                    )
                else:
                    # Delete the filter if the user is not subscribing to it
                    db.cursor().execute(
                        'DELETE FROM "filters" WHERE "id"=?',
                        (new_filter.id,)
                    )

        db.commit()
        db.close()
        os.remove(tmp_index)

        # Write new version in the repo version file
        with open(os.path.realpath(Path.REPO_VERSION), 'wt') as f:
            f.write(data.attrib["version"])
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Remote repository updated successfully')

        if warns:
            for warn in warns:
                print(f'{Fore.YELLOW}WARNING: the filter "{warn}" could not be updated as a '
                      f'custom filter takes its ID{Style.RESET_ALL}')

    unlock_database()



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Standard libraries
import os.path
import sys

# External libraries
from colorama import Fore, Style

# Local libraries
from .config import Path, Var, PROTECTING_STATUS, Font


def update_hosts(all_rules: dict) -> None:
    """Update the hosts file with a set of rules

    Args:
        all_rules (dict): Dictionary returned by tblock.rules.get_all_rules()
    """
    print(f'{Font.BOLD}==> Updating hosts file{Font.DEFAULT}')
    new_hosts = '\n\n# BEGIN BLOCKING RULES\n'
    all_rules_count = len(all_rules.keys())
    count = 1
    load_character = '|'
    for domain in all_rules:
        if count < len(all_rules):
            print(f' [{Fore.YELLOW}{load_character}{Style.RESET_ALL}] Retrieving rules '
                  f'({count}/{all_rules_count})', end='\r')
            # Show process
            if load_character == '|' and not count % 500:
                load_character = '/'
            elif load_character == '/' and not count % 500:
                load_character = '-'
            elif load_character == '-' and not count % 500:
                load_character = '\\'
            elif load_character == '\\' and not count % 500:
                load_character = '|'
        else:
            print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Retrieving rules ({count}/{all_rules_count})')
        if all_rules[domain]['policy'] == 'block':
            new_hosts += f'{Var.DEFAULT_IP}    {domain}\n'
        elif all_rules[domain]['policy'] == 'redirect':
            new_hosts += f'{all_rules[domain]["ip"]}    {domain}\n'
        count += 1
    if not os.path.isfile(Path.HOSTS_BACKUP):
        print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Backing up previous hosts file', end='\r')
        with open(Path.HOSTS, 'rt') as h:
            default_hosts_content = h.read()
        with open(Path.HOSTS_BACKUP, 'wt') as h:
            h.write(default_hosts_content)
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Backing up previous hosts file')
    else:
        with open(Path.HOSTS_BACKUP, 'rt') as h:
            default_hosts_content = h.read()
    new_hosts = default_hosts_content + new_hosts
    new_hosts = '# This file is automatically generated by TBlock\n' \
                '# To edit it manually, run "tblock -D" first. Otherwise, all your modifications will be ' \
                'overwritten.\n\n' + new_hosts
    with open(Path.HOSTS_VERIFICATION, 'wt') as f:
        f.write(new_hosts)
    with open(Path.HOSTS, 'wt') as f:
        f.write(new_hosts)
    if not PROTECTING_STATUS:
        print(
            f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Protection is now {Fore.GREEN}enabled{Style.RESET_ALL}'
        )
    print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Hosts file successfully updated')


def restore_hosts() -> None:
    print(f'{Font.BOLD}==> Restoring previous hosts file{Font.DEFAULT}')
    print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Restoring default hosts file', end='\r')
    try:
        with open(Path.HOSTS_BACKUP, 'rt') as f:
            default_hosts_content = f.read()
        with open(Path.HOSTS, 'wt') as f:
            f.write(default_hosts_content)
    except (FileNotFoundError, PermissionError):
        print(f' [{Fore.RED}x{Style.RESET_ALL}] Restoring default hosts file')
        sys.exit(f'{Fore.RED}ERROR: default hosts file already restored{Style.RESET_ALL}')
    else:
        print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Restoring default hosts file')
    print(f' [{Fore.YELLOW}|{Style.RESET_ALL}] Removing previous hosts file backup', end='\r')
    os.remove(Path.HOSTS_BACKUP)
    print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Removing previous hosts file backup')
    print(f' [{Fore.GREEN}{Font.CHECK_MARK}{Style.RESET_ALL}] Protection is now {Fore.RED}disabled{Style.RESET_ALL}')



#!/usr/bin/env python
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Import modules

import os.path
import sys
from tblock.__init__ import __version__

# Check if setuptools is installed

try:
    from setuptools import setup
except ImportError:
    setup = None
    sys.exit("Error: setuptools is not installed\nTry to run python -m pip install setuptools --upgrade")

# Open README to define long description

with open(os.path.join(os.path.dirname(__file__), "README.md"), "rt") as readme:
    long_description = readme.read()

    
# Open requirements.txt to define requirements

requirements = []

with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "rt") as f:
    for requirement in f.readlines():
        if requirement != "\n" and requirement[0:1] != "#":
            requirements.append(requirement.split("\n")[0])

setup(
        name="tblock",
        version=__version__,
        description="An anticapitalist ad-blocker that uses the hosts file",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://tblock.codeberg.page/",
        author="Twann",
        author_email="twann@ctemplar.com",
        license="GPLv3",
        packages=["tblock", "tblock.converter"],
        install_requires=requirements,
        entry_points={"console_scripts": ["tblock = tblock.cli:run", "tblockc = tblock.cli:run_converter"]},
        classifiers=[
            "Environment :: Console",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.2",
            "Programming Language :: Python :: 3.3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Operating System :: OS Independent",
        ],
)



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


# Standard libraries
import sys
import os
import re
import time
import datetime
import random

# External libraries
import requests
from colorama import Fore, Style

# Local libraries
from .config import Path
from .exceptions import InvalidLinkError


# This is meant to see which process is locking the database and to avoid others to unlock it
DB_PID = datetime.datetime.now().strftime('%Y%m%d%H%M%S%f') + str(random.randint(0, 100))


def check_root() -> bool:
    """Check if the user is the superuser.

    Returns:
        bool: True if the user is the superuser
    """
    # Just try to open the hosts file to edit it, to check if it gets restricted by the system
    try:
        with open(Path.HOSTS, 'at') as dummy_check:
            dummy_check.close()
    except PermissionError:
        return False
    else:
        return True


def lock_database() -> None:
    """Lock the database
    """
    if os.path.isfile(Path.DB_LOCK):
        print(
            f'{Fore.YELLOW}WARNING: database locked, wait for another instance of tblock to finish.\n'
            f'You can also delete "{Path.DB_LOCK}" if you are sure this is the only instance running.{Style.RESET_ALL}'
        )
    try:
        while os.path.isfile(Path.DB_LOCK):
            time.sleep(1)
        else:
            try:
                with open(Path.DB_LOCK, 'wt') as f:
                    f.write(str(DB_PID))
            except PermissionError:
                pass
    except KeyboardInterrupt:
        sys.exit(1)


def unlock_database() -> None:
    """Unlock the database
    """
    try:
        with open(Path.DB_LOCK, 'rt') as f:
            if f.read() == str(DB_PID):
                f.close()
                os.remove(Path.DB_LOCK)
    except (FileNotFoundError, PermissionError):
        pass


def is_url(url: str) -> bool:
    """Check whether a string is a valid URL or not

    Args:
        url (str): Tĥe string to check

    Returns:
        bool: True if URL seems valid
    """
    return 'http://' in url or 'https://' in url


def prompt_user(message: str, list_of_elements: list = None) -> bool:
    """Prompt the user before executing an action

    Args:
        message (str): The message to display
        list_of_elements (list, optional): A list of elements to display
    """
    output_string = ''
    line_count = 0
    if list_of_elements:
        for item in list_of_elements:
            if line_count + len(f' {item}') >= 62:
                output_string += f'\n  {item}'
                line_count = len(f' {item}')
            else:
                output_string += f' {item}'
                line_count += len(f' {item}')
    print(f':: {message}')
    if output_string:
        print(f'\n {output_string}\n')
    try:
        answer = input(':: Are you sure to continue ? [y/n] ')
    except KeyboardInterrupt:
        return False
    else:
        return answer.lower() == 'y'


def is_valid_ip(ip: str) -> bool:
    """Check whether an IP address is valid or not

    Args:
        ip (str): Tĥe IP address to check

    Returns:
        bool: True if IP address seems valid
    """
    if 7 <= len(ip) <= 15:
        if len(re.findall(r"\.", ip)) == 3:
            return True
        else:
            return False
    else:
        return False


def download_file(url: str, output_path: str) -> None:
    """Get the content of an online file and export it into a local file

    Args:
        url (str): The URL of the online file
        output_path (str): Where to export file content
    """
    online_file = requests.get(url, headers={'User-agent': 'Mozilla/5.0'})
    if online_file.status_code != 200:
        raise InvalidLinkError()
    with open(os.path.relpath(output_path), 'wb') as f:
        f.write(online_file.content)


def extract_domain(url: str) -> str:
    """Extract the domain from an URL

    Args:
        url (str): The URL

    Returns:
        str: The domain
    """
    if re.match(re.compile('.*://'), url):
        url = url.split('://')[1]
    return url.split('/')[0].split(':')[0].split('#')[0].split('?')[0]



# -*- coding: utf-8 -*-
# TBlock - An anticapitalist ad-blocker that uses the hosts file
# Copyright (C) 2021 Twann <twann@ctemplar.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class TBlockError(IOError):
    def __init__(self, *args):
        super(TBlockError, self).__init__(*args)


class SIGTERM(IOError):
    def __init__(self, *args):
        super(SIGTERM, self).__init__(*args)


class InvalidFilterSyntax(TBlockError):
    def __init__(self, *args):
        super(InvalidFilterSyntax, self).__init__(*args)


class FilterExistsError(TBlockError):
    def __init__(self, *args):
        super(FilterExistsError, self).__init__(*args)


class AlreadySubscribingError(TBlockError):
    def __init__(self, *args):
        super(AlreadySubscribingError, self).__init__(*args)


class NotSubscribingError(TBlockError):
    def __init__(self, *args):
        super(NotSubscribingError, self).__init__(*args)


class FilterNotExists(TBlockError):
    def __init__(self, *args):
        super(FilterNotExists, self).__init__(*args)


class FilterSourceExistsError(TBlockError):
    def __init__(self, *args):
        super(FilterSourceExistsError, self).__init__(*args)


class FilterNotCustomError(TBlockError):
    def __init__(self, *args):
        super(FilterNotCustomError, self).__init__(*args)


class NetworkError(TBlockError):
    def __init__(self, *args):
        super(NetworkError, self).__init__(*args)


class MissingArgumentError(TBlockError):
    def __init__(self, *args):
        super(MissingArgumentError, self).__init__(*args)


class InvalidLinkError(TBlockError):
    def __init__(self, *args):
        super(InvalidLinkError, self).__init__(*args)


class InvalidBindAddress(TBlockError):
    def __init__(self, *args):
        super(InvalidBindAddress, self).__init__(*args)


class RuleExistsError(TBlockError):
    def __init__(self, *args):
        super(RuleExistsError, self).__init__(*args)


class RuleNotExistsError(TBlockError):
    def __init__(self, *args):
        super(RuleNotExistsError, self).__init__(*args)


class InvalidRulePolicy(TBlockError):
    def __init__(self, *args):
        super(InvalidRulePolicy, self).__init__(*args)


class FilterReservedError(TBlockError):
    def __init__(self, *args):
        super(FilterReservedError, self).__init__(*args)
